/**
 * Manages buffer descriptors.
 */

#include <stdlib.h>
#include <stdio.h>
#include "BufferDescriptorManager.h"

#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>

#include "Heaps.h"

#define NOT_INITIALIZED_BUFFER (-1)
#define UNUSED_BUFFER          (0)

typedef struct {
    XDAS_Int32 id;
    XDM_BufDesc descriptor;
    XDAS_Int32 referenceCounter;
} Buffer;

typedef struct {
    int numberOfDescriptorBuffers;
    XDAS_Int32 bufferSizes[XDM_MAX_IO_BUFFERS];
    int buffersCount;
    Buffer* buffers;
} BufferDescriptorManagerInternal;

BufferDescriptorManagerHandle BufferDescriptorManager_create(int buffersCount, XDAS_Int32 numberOfDescriptorBuffers, XDAS_Int32* bufferSizes)
{
    int i;
    BufferDescriptorManagerInternal* manager = NULL;

    if ((numberOfDescriptorBuffers <= 0) || (numberOfDescriptorBuffers > XDM_MAX_IO_BUFFERS)) {
        goto FAIL;
    }

    manager = (BufferDescriptorManagerInternal*) malloc(sizeof(BufferDescriptorManagerInternal));

    if (manager == NULL) { // Failed to allocate
        goto FAIL;
    }

    manager->buffersCount = buffersCount;

    Error_Block eb;
    Error_init(&eb);

    manager->buffers = Memory_alloc(ddr3Heap, sizeof(Buffer) * buffersCount, L2_CACHE_LINE_SIZE, &eb);

    if (manager->buffers == NULL) {
        goto FAIL;
    }

    for (i = 0; i < buffersCount; i++) {
        manager->buffers[i].referenceCounter = NOT_INITIALIZED_BUFFER;
    }

    memcpy(manager->bufferSizes, bufferSizes, numberOfDescriptorBuffers * sizeof(bufferSizes[0]));
    manager->numberOfDescriptorBuffers = numberOfDescriptorBuffers;

    return (BufferDescriptorManagerHandle) manager;

    FAIL:

    if (manager != NULL) {
        BufferDescriptorManager_delete(manager);
    }

    return NULL;
}

void BufferDescriptorManager_delete(BufferDescriptorManagerHandle manager)
{
    int i;
    BufferDescriptorManagerInternal* managerInternal = (BufferDescriptorManagerInternal*) manager;
    int buffersCount = managerInternal->buffersCount;

    // Find first not initialized or unused buffer descriptor

    for (i = 0; i < buffersCount; i++) {
        Buffer* buffer = &managerInternal->buffers[i];

        if (buffer->referenceCounter != NOT_INITIALIZED_BUFFER) {
            int j;
            for (j = 0; j < managerInternal->numberOfDescriptorBuffers; j++) {
                if (buffer->descriptor.bufs[j] != NULL) {
                    Memory_free(ddr3Heap, buffer->descriptor.bufs[j], managerInternal->bufferSizes[j]);
                }
            }

            free(buffer->descriptor.bufs);
        }
    }

    Memory_free(ddr3Heap, managerInternal->buffers, sizeof(Buffer) * managerInternal->buffersCount);

    free(managerInternal);
}

XDM_BufDesc* BufferDescriptorManager_getUnusedBufferDescriptor(BufferDescriptorManagerHandle manager, XDAS_Int32 bufferDescriptorId)
{
    int i;

    BufferDescriptorManagerInternal* managerInternal = (BufferDescriptorManagerInternal*) manager;
    int buffersCount = managerInternal->buffersCount;

    // Find first not initialized or unused buffer descriptor

    for (i = 0; i < buffersCount; i++) {
        Buffer* buffer = &managerInternal->buffers[i];

        if (buffer->referenceCounter == UNUSED_BUFFER) {
            // Unused buffer already initialized - just return it

            buffer->id = bufferDescriptorId;
            buffer->referenceCounter = 1;

            return &buffer->descriptor;
        } else if (buffer->referenceCounter == NOT_INITIALIZED_BUFFER) {
            int j;

            // Initialize buffer descriptor

            int numberOfBuffers =  managerInternal->numberOfDescriptorBuffers;
            XDAS_Int8** buffers = calloc(numberOfBuffers, sizeof(XDAS_Int8*));

            if (buffers == NULL) { // Failed to allocate
                puts("Failed to allocate buffer pointers array");
                return NULL;
            }

            Error_Block eb;
            Error_init(&eb);

            for (j = 0; j < numberOfBuffers; j++) {
                 buffers[j] = Memory_alloc(ddr3Heap, managerInternal->bufferSizes[j], L2_CACHE_LINE_SIZE, &eb);

                 if (buffers[j] == NULL) {
                     free(buffers);
                     puts("Failed to allocate buffer");
                     return NULL;
                 }
            }

            buffer->descriptor.bufSizes = managerInternal->bufferSizes;
            buffer->descriptor.numBufs = numberOfBuffers;
            buffer->descriptor.bufs = buffers;
            buffer->id = bufferDescriptorId;
            buffer->referenceCounter = 1;

            return &buffer->descriptor;
        }
    }

    puts("All buffer descriptors are used");
    return NULL; // All buffer descriptors are used
}

XDAS_Int32 BufferDescriptorManager_releaseBufferDescriptor(BufferDescriptorManagerHandle manager, XDAS_Int32 bufferDescriptorId)
{
    int i;
    BufferDescriptorManagerInternal* managerInternal = (BufferDescriptorManagerInternal*) manager;
    int buffersCount = managerInternal->buffersCount;

    // Find buffer descriptor by ID and mark it as unused

    for (i = 0; i < buffersCount; i++) {
        Buffer* buffer = &managerInternal->buffers[i];

        if (buffer->id == bufferDescriptorId) {
            if (buffer->referenceCounter == 1) { // Released last reference to the buffer
                buffer->referenceCounter = UNUSED_BUFFER;
            } else { // Decrementing reference counter of the buffer
                buffer->referenceCounter--;
            }

            return 0;
        }
    }

    return 1; // No buffer descriptor with this ID
}

XDM_BufDesc* BufferDescriptorManager_getBufferDescriptorById(BufferDescriptorManagerHandle manager, XDAS_Int32 bufferDescriptorId)
{
    int i;
    BufferDescriptorManagerInternal* managerInternal = (BufferDescriptorManagerInternal*) manager;
    int buffersCount = managerInternal->buffersCount;

    // Find buffer descriptor by ID

    for (i = 0; i < buffersCount; i++) {
        Buffer* buffer = &managerInternal->buffers[i];

        if ((buffer->id == bufferDescriptorId) && (buffer->referenceCounter > UNUSED_BUFFER)) {
            return &buffer->descriptor;
        }
    }

    return NULL; // No buffer descriptor with this ID
}

int BufferDescriptorManager_referenceBufferDescriptorById(BufferDescriptorManagerHandle manager, XDAS_Int32 bufferDescriptorId)
{
    int i;
    BufferDescriptorManagerInternal* managerInternal = (BufferDescriptorManagerInternal*) manager;
    int buffersCount = managerInternal->buffersCount;

    // Find buffer descriptor by ID

    for (i = 0; i < buffersCount; i++) {
        Buffer* buffer = &managerInternal->buffers[i];

        if ((buffer->id == bufferDescriptorId) && (buffer->referenceCounter > UNUSED_BUFFER)) {
            buffer->referenceCounter++;
            return 0;
        }
    }

    return 1; // No buffer descriptor with this ID
}
