/**
 * Configuration functionality.
 */

#include "Configuration.h"

static void setDefaultValues(H264HPEncoderConfiguration *h264hp);
static void configureMainEncoder(H264HPEncoderConfiguration *h264hp);

static void configureDecoder(int decoderIndex)
{
    DspDecoderConfiguration *decoderConfiguration = &((DspConfiguration*) CONFIGURATION_ADDRESS_ON_DSP)->decoderConfigurations[decoderIndex];

    decoderConfiguration->coreIds[0] = 0;
    decoderConfiguration->coreIds[1] = 1;
    decoderConfiguration->coreIds[2] = -1;
    decoderConfiguration->decoderType = H264HP_DECODER;
    decoderConfiguration->inputBufferAddress = decoderIndex;
    decoderConfiguration->inputFrameRate = 24;
    decoderConfiguration->inputWidth = 320;
    decoderConfiguration->inputHeight = 180;
}

static void configureEncoder(int encoderIndex, int decoderIndex, int coreId)
{
    DspEncoderConfiguration *encoderConfiguration = &((DspConfiguration*) CONFIGURATION_ADDRESS_ON_DSP)->encoderConfigurations[encoderIndex];
    DspDecoderConfiguration *decoderConfiguration = &((DspConfiguration*) CONFIGURATION_ADDRESS_ON_DSP)->decoderConfigurations[decoderIndex];

    encoderConfiguration->coreIds[0] = coreId;
    encoderConfiguration->coreIds[1] = -1;

    encoderConfiguration->encoderType = H264HP_ENCODER;
    encoderConfiguration->inputDecoderIndex = decoderIndex;
    encoderConfiguration->outputBufferAddress = encoderIndex;

    H264HPEncoderConfiguration *h264hp = &encoderConfiguration->configuration.h264HP;

    setDefaultValues(&encoderConfiguration->configuration.h264HP);
    configureMainEncoder(&encoderConfiguration->configuration.h264HP);

    encoderConfiguration->outputFrameRate = decoderConfiguration->inputFrameRate;
    encoderConfiguration->outputWidth = decoderConfiguration->inputWidth;
    encoderConfiguration->outputHeight = decoderConfiguration->inputHeight;
    h264hp->bitRate = 1276000;
}

static void configureMainEncoder(H264HPEncoderConfiguration *h264hp)
{
    h264hp->entropyCodingMode = 1; // 0: CAVLC, 1: CABAC (for HP and MP only)
    h264hp->gopStructure = 0; // 0: NONUNIFORM = IBBPBBP, 1: UNIFORM = BBIBBPBB
    h264hp->IDRFrameInterval = 2; // 0: Only first I frame as IDR, 1: All I frames are IDR, 2: 1 out of 2 I frames are IDR starting from first I frame
    h264hp->transformBlockSize = 0; // Transform blocks size. 0: 4x4, 1: 8x8 (valid for HP only), 2: adaptive - encoder decides as per content
    h264hp->log2MaxFNumMinus4 = 0; // Limits the maximum frame number in the bit-stream to (1<< (log2MaxFNumMinus4 + 4)) Range[0,12]
    h264hp->picOrderCountType = 0; // POC type: 0, 1, 2
    h264hp->interlaceCodingType = 4; // 0: PICAFF, 1: MBAFF, 2: FIELDONLY_MRF, 3: FIELDONLY_ARF, 4: FIELDONLY_SPF (adaptive)
    h264hp->intraFrameInterval = 24; // The number of frames between two I frames. 0: IPPPP..., 1: IIII..., 2: IPIPIPIPI, 3: IPPIPPIPPI or IPBIPBIPBI, etc.
    h264hp->mvAccuracy = 2; // Motion vectors accuracy. 0: integer pel., 2: quarter pel.

    // General video encoder settings
    h264hp->encodingPreset = 3; // 0: DEFAULT, 1: HIGH_QUALITY, 2: HIGH_SPEED, 3: USER_DEFINED. If not USER_DEFINED most of configuration parameters are chosen by encoder, thus ignored.
    h264hp->maxInterFrameInterval = 4; // I to P frame distance. 1: no B frames, 2: one B frame, 3: two B frames, etc [1, 255]
    h264hp->profile = 77; // 66: BASELINE, 77: MAIN, 100: HIGH
    h264hp->level = 30; // 9: 1b, 10: 1.0, 11: 1.1, 12, 13, 20: 2.0, 21, 22, 30, 31, 32, 40, 41, 42, 50, 51

    // Rate control
    h264hp->rateControlPreset = 5; // 1: LOW_DELAY,  2: STORAGE, 4: NONE,  5: USER_DEFINED
    h264hp->rcAlgo = 0; // Rate control algorithm used, 0: Variable Bitrate, 1 : Constant bitrate (low dealy), 2: Constrained CBR (No Frame Skips). Useful only if rateControlPreset is set to USER_DEFINED
    h264hp->qpI = -1; // Initial QP for I/IDR frames, -1 indicates codec chosen
    h264hp->qpP = -1; // Initial QP for P frames, -1 indicates codec chosen
    h264hp->qpOffsetB = 4; // Offset of B frames QP from P frames
    h264hp->qpMaxI = 40; // Maximum QP for I/IDR frames
    h264hp->qpMinI = 14; // Minimum QP for I/IDR frames
    h264hp->qpMaxP = 51; // Maximum QP for P frames
    h264hp->qpMinP = 14; // Minimum QP for P frames
    h264hp->qpMaxB = 51; // Maximum QP for B frames
    h264hp->qpMinB = 15; // Minimum QP for B frames
    h264hp->CbQPIndexOffset = 0; // Specifies offset to be added to luma QP for addressing QPC values table for Cb components
    h264hp->CrQPIndexOffset = 0; // Specifies offset to be added to luma QP for addressing QPC values table for Cr components
    h264hp->initialBufferLevel = 2552000; // Initial Buffer level for HRD compliance, keep same as HRDBufferSize for best quality
    h264hp->HRDBufferSize = 2552000; // Hypothetical Reference Decoder Buffer Size in bits - 2*bitrate for VBR

    h264hp->enablePRC = 1; // Control Flag to enable MB level Perceptual Rate Control, 0 => Disable, 1 => Enable
    h264hp->frameSkipAfterSceneChange = 1; // 0=> no forced skip after scenechange, 1 => force skip frame after coding scene change frame.
    h264hp->scalingMatrixPreset = 0; // Scaling Matrix selection of encoder, 0: no scaling matrices, 1: normal, 2: noisy, 3 : standard default, 4 : user defined

    // Inter coding
    h264hp->interCodingPreset = 1; // Inter coding mode preset, 0 => deafult values, 1 => user defined
     h264hp->MvRangeVerP = 32; // Vertical Motion Vector Range for P frames in integer pixels, Range supported is 16 to 496
    h264hp->MvRangeHorP = 144; // Horizontal Motion Vector Range for P frames in integer pixels, Range supported is 16 to 496
    h264hp->MvRangeVerB = 32; // Vertical Motion Vector Range for B frames in integer pixels, Range supported is 16 to 496
    h264hp->MvRangeHorB = 144; // Horizontal Motion Vector Range for B frames in integer pixels, Range supported is 16 to 496
    h264hp->maxMVperMB = 4; // Maximum MV per MB (Values of 1 & 4 are valid)

    // Intra coding
    h264hp->intraCodingPreset = 1; // Intra coding mode preset, 0 => deafult values, 1 => user defined
    h264hp->enableIntraPartition = 3; // 0 => INTRA_PARTITION_NONE , 1 => INTRA_PARTITION_ISLICES, 2 => INTRA_PARTITION_IPSLICES, 3 => INTRA_PARTITION_IBSLICES, 4 => INTRA_PARTITION_IPBSLICES.
    h264hp->intraRefreshMethod = 1; // IH264_INTRAREFRESH_NONE  = 0, IH264_INTRAREFRESH_CYCLIC_MBS = 1
    h264hp->intraRefreshRate = 0; // Rate at which intra MB Refresh is done. e.g. 10 indicates every 10th MB will coded as Intra in inter pictures.
    h264hp->constrainedIntraPredEnable = 0; // Controls the intra macroblock coding in inter slices

    // Slice coding
    h264hp->sliceCodingPreset = 0; // Preset value for slice coding mode, 0 => deafult values, 1 => user defined
    h264hp->streamFormat = 0; // Type of bitstream to be encoded. Only value 0 is supprted (Byte stream format)
    h264hp->sliceMode = 1; // Type of slice coding, 0 => slice coding mode is frame based, 1 => Slices are controlled based upon number of Macroblocks
    h264hp->sliceUnitSize = 9600; // Number of macroblocks per slice

    // Loop filter
    h264hp->loopfilterPreset = 0; // Preset value for loop filter operation parameters, 0 => default values, 1 => user defined
    h264hp->loopfilterDisableIDC = 2; // Disable H.264 loop filter, (0=Filter, 1= NoFilter, 2 = No loop filter across slices)
    h264hp->filterOffsetA = 0; // Alpha offset for loop filter
    h264hp->filterOffsetB = 0; // Beta offset for loop filter

    // VUI coding
    h264hp->vuiCodingPreset = 1; // Preset value for VUI Control Params, 0 => deafult values, 1 => user defined
    h264hp->aspectRatioInfoPresentFlag = 1; // Controls the insertion of aspect ratio information in VUI part of bit-stream
    h264hp->aspectRatioIdc = 255; // Aspect ratio ID
    h264hp->videoSignalTypePresentFlag = 0; // controls the insertion of video signal type in VUI part of bit-stream
    h264hp->videoFormat = 2; // Video signal type
    h264hp->videoFullRangeFlag = 0; // Flag to specify Range of the pixels
    h264hp->colourDescriptionPresentFlag = 1; // Specifies whether colour_primaries, transfer_characteristics and matrix_coefficients are present or not.
    h264hp->colourPrimaries = 5; // chromaticity coordinates of the source primaries(Table E-3 in standard)
    h264hp->transferCharacteristics = 5; // opto-electronic transfer characteristic of the source picture(Table E-4 in standard)
    h264hp->matrixCoefficients = 5; // matrix coefficients used in deriving Y,Cb,Cr signals from the G,B,R primaries(Table E-5 in standard)
    h264hp->timingInfoPresentFlag = 1; // Controls the insertion of timing info related parameters in VUI part of bit-stream
}

static void setDefaultValues(H264HPEncoderConfiguration *h264hp)
{
    // General H.264 encoder settings
    h264hp->entropyCodingMode = DEFAULT_CONFIG_VALUE;
    h264hp->gopStructure = DEFAULT_CONFIG_VALUE;
    h264hp->IDRFrameInterval = DEFAULT_CONFIG_VALUE;
    h264hp->transformBlockSize = DEFAULT_CONFIG_VALUE;
    h264hp->log2MaxFNumMinus4 = DEFAULT_CONFIG_VALUE;
    h264hp->picOrderCountType = DEFAULT_CONFIG_VALUE;
    h264hp->interlaceCodingType = DEFAULT_CONFIG_VALUE;
    h264hp->intraFrameInterval = DEFAULT_CONFIG_VALUE;
    h264hp->mvAccuracy = DEFAULT_CONFIG_VALUE;
    h264hp->bitRate = DEFAULT_CONFIG_VALUE;

    // General video encoder settings
    h264hp->encodingPreset = DEFAULT_CONFIG_VALUE;
    h264hp->maxInterFrameInterval = DEFAULT_CONFIG_VALUE;
    h264hp->profile = DEFAULT_CONFIG_VALUE;
    h264hp->level = DEFAULT_CONFIG_VALUE;

    // Rate control
    h264hp->rateControlPreset = DEFAULT_CONFIG_VALUE;
    h264hp->rcAlgo = DEFAULT_CONFIG_VALUE;
    h264hp->qpI = DEFAULT_CONFIG_VALUE;
    h264hp->qpP = DEFAULT_CONFIG_VALUE;
    h264hp->qpOffsetB = DEFAULT_CONFIG_VALUE;
    h264hp->qpMaxI = DEFAULT_CONFIG_VALUE;
    h264hp->qpMinI = DEFAULT_CONFIG_VALUE;
    h264hp->qpMaxP = DEFAULT_CONFIG_VALUE;
    h264hp->qpMinP = DEFAULT_CONFIG_VALUE;
    h264hp->qpMaxB = DEFAULT_CONFIG_VALUE;
    h264hp->qpMinB = DEFAULT_CONFIG_VALUE;
    h264hp->CbQPIndexOffset = DEFAULT_CONFIG_VALUE;
    h264hp->CrQPIndexOffset = DEFAULT_CONFIG_VALUE;
    h264hp->initialBufferLevel = DEFAULT_CONFIG_VALUE;
    h264hp->HRDBufferSize = DEFAULT_CONFIG_VALUE;
    h264hp->enablePRC = DEFAULT_CONFIG_VALUE;
    h264hp->frameSkipAfterSceneChange = DEFAULT_CONFIG_VALUE;
    h264hp->scalingMatrixPreset = DEFAULT_CONFIG_VALUE;

    // Inter coding
    h264hp->interCodingPreset = DEFAULT_CONFIG_VALUE;
    h264hp->MvRangeVerP = DEFAULT_CONFIG_VALUE;
    h264hp->MvRangeHorP = DEFAULT_CONFIG_VALUE;
    h264hp->MvRangeVerB = DEFAULT_CONFIG_VALUE;
    h264hp->MvRangeHorB = DEFAULT_CONFIG_VALUE;
    h264hp->maxMVperMB = DEFAULT_CONFIG_VALUE;

    // Intra coding
    h264hp->intraCodingPreset = DEFAULT_CONFIG_VALUE;
    h264hp->enableIntraPartition = DEFAULT_CONFIG_VALUE;
    h264hp->intraRefreshMethod = DEFAULT_CONFIG_VALUE;
    h264hp->intraRefreshRate = DEFAULT_CONFIG_VALUE;
    h264hp->constrainedIntraPredEnable = DEFAULT_CONFIG_VALUE;

    // Slice coding
    h264hp->sliceCodingPreset = DEFAULT_CONFIG_VALUE;
    h264hp->streamFormat = DEFAULT_CONFIG_VALUE;
    h264hp->sliceMode = DEFAULT_CONFIG_VALUE;
    h264hp->sliceUnitSize = DEFAULT_CONFIG_VALUE;

    // Loop filter
    h264hp->loopfilterPreset = DEFAULT_CONFIG_VALUE;
    h264hp->loopfilterDisableIDC = DEFAULT_CONFIG_VALUE;
    h264hp->filterOffsetA = DEFAULT_CONFIG_VALUE;
    h264hp->filterOffsetB = DEFAULT_CONFIG_VALUE;

    // VUI coding
    h264hp->vuiCodingPreset = DEFAULT_CONFIG_VALUE;
    h264hp->aspectRatioInfoPresentFlag = DEFAULT_CONFIG_VALUE;
    h264hp->aspectRatioIdc = DEFAULT_CONFIG_VALUE;
    h264hp->videoSignalTypePresentFlag = DEFAULT_CONFIG_VALUE;
    h264hp->videoFormat = DEFAULT_CONFIG_VALUE;
    h264hp->videoFullRangeFlag = DEFAULT_CONFIG_VALUE;
    h264hp->colourDescriptionPresentFlag = DEFAULT_CONFIG_VALUE;
    h264hp->colourPrimaries = DEFAULT_CONFIG_VALUE;
    h264hp->transferCharacteristics = DEFAULT_CONFIG_VALUE;
    h264hp->matrixCoefficients = DEFAULT_CONFIG_VALUE;
    h264hp->timingInfoPresentFlag = DEFAULT_CONFIG_VALUE;
}

const DspConfiguration* Configuration_get()
{
    DspConfiguration* configuration = (DspConfiguration*) CONFIGURATION_ADDRESS_ON_DSP;

    if (configuration->marker != 0xABBA) {
        configuration->marker = 0xABBA;

        configuration->decodersCount = 1;
        configureDecoder(0);

        configuration->encodersCount = 1;
        configureEncoder(0, 0, 2);
    }

    return configuration;
}

int Configuration_calculateNumberOfCores(const int32_t* coreIds)
{
    int i;
    int count = 0;

    for (i = 0; i < CORES_PER_DSP; i++) {
        if (coreIds[i] == -1) {
            break;
        }
        count++;
    }

    return count;
}
