/**
 * Master task that encodes.
 */

#include <stdio.h>

#include <xdc/std.h>
#include <ti/sysbios/knl/Clock.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/family/c66/Cache.h>
#include <ti/sdo/codecs/h264hpvenc/h264hpvenc_ti.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>

#include "H264HighProfileEncoder.h"
#include "BufferDescriptorManager.h"
#include "Heaps.h"
#include "Configuration.h"
#include "GlobalGates.h"
#include "MulticoreApi.h"

#define TRANSFER_YUV  (0)

#define INPUT_DATA_WAIT_INTERVAL (10000 / Clock_tickPeriod) // 10 milliseconds

#include "Heaps.h"

static ResultCode doMasterEncoding(int encoderIndex);

void encodeMasterTask(UArg encoderIndex, UArg arg1)
{
    const DspEncoderConfiguration* const encoderConfiguration = &(Configuration_get()->encoderConfigurations[encoderIndex]);
    const int numberOfEncoderCores = Configuration_calculateNumberOfCores(encoderConfiguration->coreIds);

    while (1) {
        ResultCode error = doMasterEncoding(encoderIndex);

        switch (error) {
        case ResultCode_Error:
            puts("Fatal master encoder task error. Terminating task.");
            goto TASK_END;

        case ResultCode_ErrorNonFatal:
            puts("Non fatal master encoder task error. Restarting task.");
            break;

        default:
            puts("Unknown master encoder task error. Terminating task.");
            goto TASK_END;
        }

        fflush(stdout);
    }

TASK_END:

    fflush(stdout);

    printf("Master encoder #%d task terminated\n", encoderIndex);
    fflush(stdout);
}

static ResultCode doMasterEncoding(int encoderIndex)
{
    int i;
    int result;
    ResultCode resultError = ResultCode_Error;
    const DspConfiguration* configuration = Configuration_get();
    const DspEncoderConfiguration* const encoderConfiguration = &configuration->encoderConfigurations[encoderIndex];
    Encoder2* encoder = NULL;
    IVIDENC2_Handle encoderHandle = NULL;
    IVIDENC2_Fxns* encoderFunctions = NULL;
    IVIDENC2_InArgs* inEncoderArguments = NULL;
    IVIDENC2_OutArgs* outEncoderArguments = NULL;
    IVIDENC2_DynamicParams* encoderDynamicParams = NULL;
    IVIDENC2_Status* encoderStatus = NULL;
    IALG_MemRec* memTabs;
    int outEncoderBufferSize = 0;
    Error_Block eb;
    int8_t* outPointer;

    XDM2_BufDesc outEncoderBufferDescriptor = { 0 };
    IVIDEO2_BufDesc inEncoderBufferDescriptor = { 0 };

    // Shared data, synchronization, multi-core
    int algorithmIndex = Common_getEncoderAlgorithmIndex(encoderIndex);
    int numberOfEncoderCores = Configuration_calculateNumberOfCores(encoderConfiguration->coreIds);

    Error_init(&eb);

    printf("Initializing master encoder %d\n", encoderIndex);

    encoder = H264HighProfileEncoder_create(encoderConfiguration, encoderIndex);

    if (encoder == NULL) {
        puts("Failed to create encoder");
        goto TASK_END;
    }

    puts("Encoder created");
    fflush(stdout);

    encoderFunctions = encoder->functions;
    encoderHandle = encoder->handle;
    inEncoderArguments = encoder->inArguments;
    outEncoderArguments = encoder->outArguments;
    encoderDynamicParams = encoder->dynamicParams;
    encoderStatus = encoder->status;
    memTabs = encoder->memTabs;

    puts("Configure encoder buffers");

    encoder->activate(encoder);
    result = encoderFunctions->control(encoderHandle, XDM_GETBUFINFO, encoderDynamicParams, encoderStatus);
    encoder->deactivate(encoder);

    if (result != IVIDENC2_EOK) {
        printf("XDM_GETBUFINFO failed (%d)\n", result);
        goto TASK_END;
    }

    printf("minNumInBufs %d\n"
            "minNumOutBufs %d\n", encoderStatus->bufInfo.minNumInBufs, encoderStatus->bufInfo.minNumOutBufs);

    for (i = 0; i < encoderStatus->bufInfo.minNumInBufs; i++) {
        printf("minInBufSize[%d] %d %d\n", i, encoderStatus->bufInfo.minInBufSize[i].bytes, encoderStatus->bufInfo.inBufMemoryType[i]);
    }

    for (i = 0; i < encoderStatus->bufInfo.minNumOutBufs; i++) {
        printf("minOutBufSize[%d] %d %d\n", i, encoderStatus->bufInfo.minOutBufSize[i].bytes, encoderStatus->bufInfo.outBufMemoryType[i]);
    }

    // Allocate output encoder buffers

    outEncoderBufferSize = 0;

    for (i = 0; i < encoderStatus->bufInfo.minNumOutBufs; i++) {
        outEncoderBufferSize += encoderStatus->bufInfo.minOutBufSize[i].bytes;
    }

    outPointer = Memory_alloc(ddr3Heap, outEncoderBufferSize, L2_CACHE_LINE_SIZE, &eb);

    if (outPointer == NULL) {
        puts("Failed to allocate output encoder buffer");
        goto TASK_END;
    }

    outEncoderBufferDescriptor.numBufs = encoderStatus->bufInfo.minNumOutBufs;

    for (i = 0; i < encoderStatus->bufInfo.minNumOutBufs; i++) {
        XDM2_SingleBufDesc *bufferDescriptor = outEncoderBufferDescriptor.descs + i;
        bufferDescriptor->bufSize = encoderStatus->bufInfo.minOutBufSize[i];
        bufferDescriptor->memType = encoderStatus->bufInfo.outBufMemoryType[i];
        bufferDescriptor->buf = outPointer;

        outPointer += bufferDescriptor->bufSize.bytes;
    }

    // Configure input encoder buffer parameters

    inEncoderBufferDescriptor.dataLayout = IVIDEO_FIELD_INTERLEAVED;
    inEncoderBufferDescriptor.topFieldFirstFlag = XDAS_TRUE;
    inEncoderBufferDescriptor.numPlanes = encoderStatus->bufInfo.minNumInBufs;
    inEncoderBufferDescriptor.contentType = encoder->params->inputContentType;
    inEncoderBufferDescriptor.chromaFormat = encoder->params->inputChromaFormat;

    if (encoderStatus->bufInfo.minNumInBufs != 3) {
        puts("Only 3 YUV planes supported");
        goto TASK_END;
    }

    for (i = 0; i < encoderStatus->bufInfo.minNumInBufs; i++) {
        inEncoderBufferDescriptor.planeDesc[i].memType = encoderStatus->bufInfo.inBufMemoryType[i];
        inEncoderBufferDescriptor.planeDesc[i].bufSize = encoderStatus->bufInfo.minInBufSize[i];
    }

    // Signal to decoder that next encoder is ready

    puts("Encoder initialization OK");

    printf("encoder %d\n", encoderIndex);
    fflush(stdout);

    inEncoderArguments->inputID = 0;

    while (1) {
        // Prepare input encoder buffer

        if (inEncoderBufferDescriptor.planeDesc[0].buf == NULL) {
            inEncoderBufferDescriptor.planeDesc[0].buf = Memory_alloc(ddr3Heap, encoderStatus->bufInfo.minInBufSize[0].bytes, L2_CACHE_LINE_SIZE, &eb);
            inEncoderBufferDescriptor.planeDesc[1].buf = Memory_alloc(ddr3Heap, encoderStatus->bufInfo.minInBufSize[1].bytes, L2_CACHE_LINE_SIZE, &eb);
            inEncoderBufferDescriptor.planeDesc[2].buf = Memory_alloc(ddr3Heap, encoderStatus->bufInfo.minInBufSize[2].bytes, L2_CACHE_LINE_SIZE, &eb);
        }

        Cache_inv(inEncoderBufferDescriptor.planeDesc[0].buf, inEncoderBufferDescriptor.planeDesc[0].bufSize.bytes, Cache_Type_ALL, TRUE);
        Cache_inv(inEncoderBufferDescriptor.planeDesc[1].buf, inEncoderBufferDescriptor.planeDesc[1].bufSize.bytes, Cache_Type_ALL, TRUE);
        Cache_inv(inEncoderBufferDescriptor.planeDesc[2].buf, inEncoderBufferDescriptor.planeDesc[2].bufSize.bytes, Cache_Type_ALL, TRUE);

        inEncoderArguments->inputID++;

        Cache_inv(outEncoderBufferDescriptor.descs[0].buf, outEncoderBufferSize, Cache_Type_ALL, TRUE);

        encoder->activate(encoder);
        result = encoderFunctions->process(encoderHandle, &inEncoderBufferDescriptor, &outEncoderBufferDescriptor, inEncoderArguments, outEncoderArguments);
        encoder->shmmap_sync(encoder->encoderTaskIndex, IVIDMC_SHMEM_KEY_CLEAN, NULL);
        Algorithm_synchronize(memTabs);
        encoder->deactivate(encoder);

        Cache_wbInv(outEncoderBufferDescriptor.descs[0].buf, outEncoderBufferSize, Cache_Type_ALL, TRUE);

        if (result != IVIDENC2_EOK) {
            printf("Encoder process call failed (%d %d)\n", result, outEncoderArguments->extendedError);
            resultError = ResultCode_ErrorNonFatal;
            goto TASK_END;
        }

        if (outEncoderArguments->bytesGenerated > 0) { // Output encoded frame
        }
    }

TASK_END:

    fflush(stdout);

    if (encoder != NULL) {
        encoder->delete(encoder);
    }

    if (outEncoderBufferDescriptor.descs[0].buf != NULL) {
        Memory_free(ddr3Heap, outEncoderBufferDescriptor.descs[0].buf, outEncoderBufferSize);
    }

    fflush(stdout);
    return resultError;
}
