/**
 * H.264 High Profile decoder.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <xdc/std.h>
#include <ti/sdo/fc/rman/rman.h>
#include <ti/sdo/codecs/h264hpvdec/ih264hpvdec.h>

#include "AlgorithmUtils.h"
#include "H264HighProfileDecoder.h"
#include "MulticoreApi.h"
#include "Configuration.h"
#include "Common.h"

#include <ti/sysbios/family/c66/Cache.h>

extern IH264HPVDEC_Fxns H264HPVDEC_TI_IH264HPVDEC;
extern IRES_Fxns H264HPVDEC_TI_IRES;

extern cregister volatile unsigned int DNUM;

typedef struct {
    Decoder decoder;
    IH264HPVDEC_InArgs inArguments;
    IH264HPVDEC_OutArgs outArguments;
    IH264HPVDEC_Status status;
    IH264HPVDEC_Params params;
    IH264HPVDEC_DynamicParams dynamicParams;
    IVIDMC3_t multicoreParams;
    int isResourcesAssigned;
    int width;
    int height;
} H264HighProfileDecoder;

static void H264HighProfileDecoder_activate(Decoder* decoder)
{
    MulticoreApi_setActiveAlgorithm(decoder->algorithmIndex);
    decoder->functions->ialg.algActivate((IALG_Handle) decoder->handle);
}

static void H264HighProfileDecoder_deactivate(Decoder* decoder)
{
    decoder->functions->ialg.algDeactivate((IALG_Handle) decoder->handle);
    MulticoreApi_setActiveAlgorithm(-1);
}

static ResultCode H264HighProfileDecoder_initializeMulticoreApi(H264HighProfileDecoder *decoder, const DspDecoderConfiguration *decoderConfiguration, int decoderIndex)
{
    int decoderTaskIndex = -1;
    int decodersCount = 0;

    int coreIndex;
    for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
        int coreId = decoderConfiguration->coreIds[coreIndex];

        if (coreId == -1) {
            break;
        }

        decodersCount++;

        if (coreId == DNUM) {
            if (decoderTaskIndex != -1) {
                puts("More then one DNUM in decoder configuration");
                return ResultCode_Error;
            }

            decoderTaskIndex = coreIndex;
        }
    }

    if (decoderTaskIndex == -1) {
        puts("Failed to find DNUM in decoder configuration");
        return ResultCode_Error;
    }

    decoder->decoder.algorithmIndex = Common_getDecoderAlgorithmIndex(decoderIndex);
    decoder->decoder.decoderTaskIndex = decoderTaskIndex;
    decoder->decoder.decodersCount = decodersCount;

    decoder->multicoreParams.user_id = decoderTaskIndex;
    decoder->multicoreParams.num_users = decodersCount;
    if (decoderTaskIndex == 0) {
        decoder->multicoreParams.task_ID = IVIDMC3_TASK_MASTER;
    } else {
        decoder->multicoreParams.task_ID = IVIDMC3_TASK_LOCMASTER;
    }

    /*printf("user_id=%d\n", decoder->multicoreParams.user_id);
    printf("num_users=%d\n", decoder->multicoreParams.num_users);
    printf("task_ID=%d\n", decoder->multicoreParams.task_ID);*/

    if (decodersCount > 1) { // Multi-core encoding
        decoder->multicoreParams.keyCreate = MulticoreApi3_keyCreate;
        decoder->multicoreParams.lockAcquire = MulticoreApi3_lockAcquire;
        decoder->multicoreParams.lockCheck = MulticoreApi3_lockCheck;
        decoder->multicoreParams.lockRelease = MulticoreApi3_lockRelease;
        decoder->multicoreParams.barrWait = MulticoreApi3_barrWait;
        decoder->decoder.swbarr = MulticoreApi_swbarrWithYield;
    } else { // Single core encoding
        decoder->multicoreParams.keyCreate = MulticoreApi3_keyCreate;
        decoder->multicoreParams.lockAcquire = MulticoreApi3_lockAcquire;
        decoder->multicoreParams.lockCheck = MulticoreApi3_lockCheck;
        decoder->multicoreParams.lockRelease = MulticoreApi3_lockRelease;
        decoder->multicoreParams.barrWait = MulticoreApi3_barrWait;
        decoder->decoder.swbarr = MulticoreApi_swbarrWithYieldDummy;
    }

    decoder->multicoreParams.shmMap = MulticoreApi3_shmMap;
    decoder->multicoreParams.shmSync = MulticoreApi3_shmSync;
    decoder->multicoreParams.shmSyncWait = MulticoreApi3_shmSyncWait;

    decoder->params.coreTeam = calloc(decodersCount, sizeof(*decoder->params.coreTeam));

    if (decoder->params.coreTeam == NULL) {
        puts("Failed to allocate core team");
        return ResultCode_Error;
    }

    //puts("core team:");
    int i;
    for (i = 0; i < decodersCount; i++) {
        //printf("core #%d\n", decoderConfiguration->coreIds[i]);
        decoder->params.coreTeam[i] = decoderConfiguration->coreIds[i];
    }

    decoder->decoder.activate = H264HighProfileDecoder_activate;
    decoder->decoder.deactivate = H264HighProfileDecoder_deactivate;

    return ResultCode_Ok;
}

static void H264HighProfileDecoder_configure(H264HighProfileDecoder *decoder, int frameRate, int width, int height)
{
    decoder->params.mcViddecParams = &decoder->multicoreParams;
    decoder->params.viddecParams.size = sizeof(decoder->params);
    decoder->dynamicParams.viddecDynamicParams.size = sizeof(decoder->dynamicParams);

    decoder->params.viddecParams.maxFrameRate = frameRate * 1000;
    decoder->params.viddecParams.maxBitRate = 10000000; // TODO: should be configurable
    decoder->params.viddecParams.dataEndianness = XDM_BYTE;
    decoder->params.viddecParams.maxWidth = (width % 16 == 0) ? width : (width / 16) * 16 + 16; // Image width in pixels, must be multiple of 16
    decoder->params.viddecParams.maxHeight = (height % 16 == 0) ? height : (height / 16) * 16 + 16; // Image height in pixels, must be multiple of 16
    decoder->params.viddecParams.forceChromaFormat = XDM_YUV_420P; // 1 => XMI_YUV_420P, 3 => XMI_YUV_422IBE, 4 => XMI_YUV_422ILE
    decoder->params.displayDelay = 16;

    if (decoder->params.viddecParams.maxWidth < 1280) {
        decoder->params.viddecParams.maxWidth = 1280;
    }

    if (decoder->params.viddecParams.maxHeight < 720) {
        decoder->params.viddecParams.maxHeight = 720;
    }

    /*printf("width=%d\n", decoder->params.viddecParams.maxWidth);
    printf("height=%d\n", decoder->params.viddecParams.maxHeight);*/

    decoder->dynamicParams.viddecDynamicParams.decodeHeader = XDM_DECODE_AU;
    decoder->dynamicParams.viddecDynamicParams.frameSkipMode = IVIDEO_NO_SKIP;
    decoder->dynamicParams.viddecDynamicParams.frameOrder = IVIDDEC2_DISPLAY_ORDER;
}

static void H264HighProfileDecoder_delete(Decoder *decoder)
{
    int result;
    H264HighProfileDecoder* h264Decoder = (H264HighProfileDecoder*) decoder;

    if (h264Decoder != NULL) {
        if (h264Decoder->isResourcesAssigned) {
            decoder->activate(decoder);
            result = RMAN_deactivateAllResources((IALG_Handle) decoder->handle, &H264HPVDEC_TI_IRES, 0);
            decoder->deactivate(decoder);
            if (result != IRES_OK) {
                printf("WARNING: Deactivate resources failed (%d)\n", result);
            }

            decoder->activate(decoder);
            result = RMAN_freeResources((IALG_Handle) decoder->handle, &H264HPVDEC_TI_IRES, 0);
            decoder->deactivate(decoder);
            if (result != IRES_OK) {
                printf("WARNING: Free resources failed (%d)\n", result);
            }
        }

        if (h264Decoder->decoder.handle != NULL) {
            MulticoreApi_setActiveAlgorithm(decoder->algorithmIndex);
            ResultCode result = Algorithm_delete((IALG_Handle) decoder->handle);
            MulticoreApi_setActiveAlgorithm(-1);
            if (result != ResultCode_Ok) {
                puts("Failed to delete decoder");
            }
        }

        if (h264Decoder->params.coreTeam != NULL) {
            free(h264Decoder->params.coreTeam);
        }

        free(h264Decoder);
    }
}

Decoder* H264HighProfileDecoder_create(const DspDecoderConfiguration *decoderConfiguration, int decoderIndex)
{
    int result;
    IVIDDEC2_Fxns* decoderFunctions;
    H264HighProfileDecoder* h264Decoder = malloc(sizeof(H264HighProfileDecoder));
    int numberOfDecoderCores = Configuration_calculateNumberOfCores(decoderConfiguration->coreIds);

    if (h264Decoder == NULL) {
        puts("Failed to allocate memory for decoder");
        goto CREATE_ERROR;
    }

    memset(h264Decoder, 0, sizeof(H264HighProfileDecoder));

    h264Decoder->decoder.inArguments = &h264Decoder->inArguments.viddecInArgs;
    h264Decoder->decoder.outArguments = &h264Decoder->outArguments.viddecOutArgs;
    h264Decoder->decoder.status = &h264Decoder->status.viddecStatus;
    h264Decoder->decoder.params = &h264Decoder->params.viddecParams;
    h264Decoder->decoder.dynamicParams = &h264Decoder->dynamicParams.viddecDynamicParams;
    h264Decoder->decoder.delete = H264HighProfileDecoder_delete;

    h264Decoder->width = decoderConfiguration->inputWidth;
    h264Decoder->height = decoderConfiguration->inputHeight;

    h264Decoder->status.viddecStatus.size = sizeof(h264Decoder->status);
    h264Decoder->inArguments.viddecInArgs.size = sizeof(h264Decoder->inArguments);
    h264Decoder->outArguments.viddecOutArgs.size = sizeof(h264Decoder->outArguments);

    H264HighProfileDecoder_configure(h264Decoder, decoderConfiguration->inputFrameRate, decoderConfiguration->inputWidth, decoderConfiguration->inputHeight);

    if (H264HighProfileDecoder_initializeMulticoreApi(h264Decoder, decoderConfiguration, decoderIndex) != ResultCode_Ok) {
        goto CREATE_ERROR;
    }

    MulticoreApi_setActiveAlgorithm(h264Decoder->decoder.algorithmIndex);

    h264Decoder->decoder.handle = (IVIDDEC2_Handle) Algorithm_create(
            &H264HPVDEC_TI_IH264HPVDEC.ividdec.ialg,
            (IALG_Params*) &h264Decoder->params,
            h264Decoder->decoder.memTabs);

    MulticoreApi_setActiveAlgorithm(-1);

    if (h264Decoder->decoder.handle == NULL) {
        goto CREATE_ERROR;
    }

    h264Decoder->decoder.functions = h264Decoder->decoder.handle->fxns;
    decoderFunctions = h264Decoder->decoder.functions;

    // Configure decoder resources

    h264Decoder->decoder.activate(&h264Decoder->decoder);
    result = RMAN_assignResources((IALG_Handle) h264Decoder->decoder.handle, &H264HPVDEC_TI_IRES, 0);

    if (result != IRES_OK) {
        h264Decoder->decoder.deactivate(&h264Decoder->decoder);
        printf("Assign resources failed (%d)\n", result);
        goto CREATE_ERROR;
    }

    h264Decoder->isResourcesAssigned = 1;

    result = RMAN_activateAllResources((IALG_Handle) h264Decoder->decoder.handle, &H264HPVDEC_TI_IRES, 0);
    h264Decoder->decoder.deactivate(&h264Decoder->decoder);

    //if (result != IRES_OK) {
        // TODO: WARNING: fix this
    //}

    // Configure decoder

    h264Decoder->decoder.activate(&h264Decoder->decoder);
    result = decoderFunctions->control(h264Decoder->decoder.handle, XDM_RESET, &h264Decoder->dynamicParams.viddecDynamicParams, &h264Decoder->status.viddecStatus);
    h264Decoder->decoder.deactivate(&h264Decoder->decoder);

    if (result != IVIDDEC2_EOK) {
        printf("XDM_RESET failed (%d) with extended error (%d)\n", result, h264Decoder->status.viddecStatus.extendedError);
        goto CREATE_ERROR;
    }

    // Set params

    h264Decoder->decoder.activate(&h264Decoder->decoder);
    result = decoderFunctions->control(h264Decoder->decoder.handle, XDM_SETPARAMS, &h264Decoder->dynamicParams.viddecDynamicParams, &h264Decoder->status.viddecStatus);
    h264Decoder->decoder.deactivate(&h264Decoder->decoder);

    if (result != IVIDDEC2_EOK) {
        printf("XDM_SETPARAMS failed (%d) with extended error (%d)\n", result, h264Decoder->status.viddecStatus.extendedError);
        goto CREATE_ERROR;
    }

    MulticoreApi_setActiveAlgorithm(h264Decoder->decoder.algorithmIndex);
    Cache_wbInvAll();
    Algorithm_synchronize(h264Decoder->decoder.memTabs);
    MulticoreApi_swbarr(DNUM, 0, numberOfDecoderCores);
    MulticoreApi_setActiveAlgorithm(-1);

    return (Decoder*) h264Decoder;

CREATE_ERROR:

    H264HighProfileDecoder_delete((Decoder*) h264Decoder);
    return NULL;
}

