/**
 * H.264 High Profile encoder.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <xdc/std.h>
#include <ti/xdais/ires.h>
#include <ti/sdo/codecs/h264hpvenc/h264hpvenc_ti.h>
#include <ti/sdo/fc/rman/rman.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>

#include "DspCommonTypes.h"
#include "AlgorithmUtils.h"
#include "H264HighProfileEncoder.h"
#include "Heaps.h"
#include "MulticoreApi.h"
#include "Configuration.h"

typedef struct {
    Encoder2 encoder;
    IH264HPVENC_InArgs inArguments;
    IH264HPVENC_OutArgs outArguments;
    IH264HPVENC_Status status;
    IVIDMC_t multicoreParams;
    IH264HPVENC_Params params;
    IH264HPVENC_DynamicParams dynamicParams;
    int isResourcesAssigned;
} H264HighProfileEncoder;

extern IRES_Fxns H264HPVENC_TI_IRES;
extern IH264HPVENC_Fxns H264HPVENC_TI_IH264HPVENC;

extern cregister volatile unsigned int DNUM;

static void H264HighProfileEncoder_activate(Encoder2* encoder)
{
    MulticoreApi_setActiveAlgorithm(encoder->algorithmIndex);
    encoder->functions->ialg.algActivate((IALG_Handle) encoder->handle);
}

static void H264HighProfileEncoder_deactivate(Encoder2* encoder)
{
    encoder->functions->ialg.algDeactivate((IALG_Handle) encoder->handle);
    MulticoreApi_setActiveAlgorithm(-1);
}

static ResultCode H264HighProfileEncoder_initializeMulticoreApi(H264HighProfileEncoder *encoder, const DspEncoderConfiguration *encoderConfiguration, int encoderIndex)
{
    int encoderTaskIndex = -1;
    int encodersCount = 0;

    int coreIndex;
    for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
        int coreId = encoderConfiguration->coreIds[coreIndex];

        if (coreId == -1) {
            break;
        }

        encodersCount++;

        if (coreId == DNUM) {
            if (encoderTaskIndex != -1) {
                puts("More then one DNUM in encoder configuration");
                return ResultCode_Error;
            }

            encoderTaskIndex = coreIndex;
        }
    }

    if (encoderTaskIndex == -1) {
        puts("Failed to find DNUM in encoder configuration");
        return ResultCode_Error;
    }

    encoder->encoder.algorithmIndex = Common_getEncoderAlgorithmIndex(encoderIndex);
    encoder->encoder.encoderTaskIndex = encoderTaskIndex;
    encoder->encoder.encodersCount = encodersCount;

    encoder->multicoreParams.coreID = encoderTaskIndex;
    encoder->multicoreParams.ncores = encodersCount;
    encoder->multicoreParams.core_task_ID = (encoderTaskIndex == 0) ? IVIDMC_TASK_MASTER : IVIDMC_TASK_SLAVE;

    encoder->multicoreParams.shmmap = MulticoreApi_shmmap;
    encoder->multicoreParams.shmmap_sync = MulticoreApi_shmmap_sync;
    encoder->multicoreParams.shmmunmap = MulticoreApi_shmmunmap;

    if (encodersCount > 1) { // Multi-core encoding
        encoder->multicoreParams.swbarr = MulticoreApi_swbarr;
        encoder->encoder.swbarr = MulticoreApi_swbarrWithYield;
    } else { // Single core encoding
        encoder->multicoreParams.swbarr = MulticoreApi_swbarrDummy;
        encoder->encoder.swbarr = MulticoreApi_swbarrWithYieldDummy;
    }

    encoder->encoder.activate = H264HighProfileEncoder_activate;
    encoder->encoder.deactivate = H264HighProfileEncoder_deactivate;
    encoder->encoder.shmmap_sync = encoder->params.ividmc->shmmap_sync;

    return ResultCode_Ok;
}

static void H264HighProfileEncoder_configure(H264HighProfileEncoder *encoder, const DspEncoderConfiguration *encoderConfiguration)
{
    const H264HPEncoderConfiguration *h264HPEncoderConfiguration = &encoderConfiguration->configuration.h264HP;

    encoder->params.ividmc = &encoder->multicoreParams;
    encoder->params.videnc2Params.size = sizeof(encoder->params);
    encoder->dynamicParams.videnc2DynamicParams.size = sizeof(encoder->dynamicParams);

#define CONFIGURE(parameterName, defaultValue) (((h264HPEncoderConfiguration->##parameterName) == DEFAULT_CONFIG_VALUE) ? defaultValue : h264HPEncoderConfiguration->##parameterName);

    encoder->params.entropyCodingMode = CONFIGURE(entropyCodingMode, IH264HP_ENTROPYCODING_CAVLC); // 0: CAVLC, 1: CABAC (for HP and MP only)
    encoder->params.gopStructure = CONFIGURE(gopStructure, IH264HP_GOPSTRUCTURE_NONUNIFORM); // 0: NONUNIFORM = IBBPBBP, 1: UNIFORM = BBIBBPBB
    encoder->params.IDRFrameInterval = CONFIGURE(IDRFrameInterval, 0); // 0: Only first I frame as IDR, 1: All I frames are IDR, 2: 1 out of 2 I frames are IDR starting from first I frame
    encoder->params.transformBlockSize = CONFIGURE(transformBlockSize, IH264HP_TRANSFORM_4x4); // Transform blocks size. 0: 4x4, 1: 8x8 (valid for HP only), 2: adaptive - encoder decides as per content
    encoder->params.log2MaxFNumMinus4 = CONFIGURE(log2MaxFNumMinus4, 0); // Limits the maximum frame number in the bit-stream to (1<< (log2MaxFNumMinus4 + 4)) Range[0,12]
    encoder->params.picOrderCountType = CONFIGURE(picOrderCountType, 0);  // POC type: 0, 1, 2
    encoder->params.interlaceCodingType = CONFIGURE(interlaceCodingType, 4); // 0: PICAFF, 1: MBAFF, 2: FIELDONLY_MRF, 3: FIELDONLY_ARF, 4: FIELDONLY_SPF (adaptive)
    encoder->params.debugTraceLevel = 0; // Not configurable
    encoder->params.lastNFramesToLog = 0; // Not configurable

    encoder->params.videnc2Params.encodingPreset                = CONFIGURE(encodingPreset, XDM_DEFAULT); // 0: DEFAULT, 1: HIGH_QUALITY, 2: HIGH_SPEED, 3: USER_DEFINED. If not USER_DEFINED most of configuration parameters are chosen by encoder, thus ignored.
    encoder->params.videnc2Params.rateControlPreset             = CONFIGURE(rateControlPreset, IVIDEO_LOW_DELAY); // 1: LOW_DELAY,  2: STORAGE, 4: NONE,  5: USER_DEFINED
    encoder->params.videnc2Params.maxWidth                      = encoderConfiguration->outputWidth;
    encoder->params.videnc2Params.maxHeight                     = encoderConfiguration->outputHeight;
    encoder->params.videnc2Params.dataEndianness                = XDM_BYTE; // Not configurable
    encoder->params.videnc2Params.maxInterFrameInterval         = CONFIGURE(maxInterFrameInterval, 1); // I to P frame distance. 1: no B frames, 2: one B frame, 3: two B frames, etc [1, 255]
    encoder->params.videnc2Params.maxBitRate                    = CONFIGURE(bitRate, 512000); // In bits per second. Should be valid as per LEVEL limit
    encoder->params.videnc2Params.minBitRate                    = 0; // In bits per second
    encoder->params.videnc2Params.inputChromaFormat             = XDM_YUV_420P; // Not configurable
    encoder->params.videnc2Params.inputContentType              = IVIDEO_PROGRESSIVE; // 0: PROGRESSIVE, 1: INTERLACED
    encoder->params.videnc2Params.operatingMode                 = IVIDEO_ENCODE_ONLY; // Not configurable
    encoder->params.videnc2Params.profile                       = CONFIGURE(profile, IH264HP_BASELINE_PROFILE); // 66: BASELINE, 77: MAIN, 100: HIGH
    encoder->params.videnc2Params.level                         = CONFIGURE(level, IH264HP_LEVEL_40); // 9: 1b, 10: 1.0, 11: 1.1, 12, 13, 20: 2.0, 21, 22, 30, 31, 32, 40, 41, 42, 50, 51
    encoder->params.videnc2Params.inputDataMode                 = IVIDEO_ENTIREFRAME; // Not configurable
    encoder->params.videnc2Params.outputDataMode                = IVIDEO_ENTIREFRAME; // Not configurable

    int i;
    for (i = 0 ; i < IVIDEO_MAX_NUM_METADATA_PLANES; i++) {
        encoder->params.videnc2Params.metadataType[i] = IVIDEO_METADATAPLANE_NONE;
    }

    // Rate control

    encoder->params.rateControlParams.rateControlParamsPreset = encoder->params.videnc2Params.rateControlPreset == IVIDEO_USER_DEFINED ? IH264HP_RATECONTROLPARAMS_USERDEFINED : IH264HP_RATECONTROLPARAMS_DEFAULT;
    encoder->params.rateControlParams.rcAlgo                  = CONFIGURE(rcAlgo, IH264HP_RATECONTROL_VBR); // Rate control algorithm used, 0: Variable Bitrate, 1 : Constant bitrate (low dealy), 2: Constrained CBR (No Frame Skips). Useful only if rateControlPreset is set to USER_DEFINED
    encoder->params.rateControlParams.qpI                     = CONFIGURE(qpI, -1); // Initial QP for I/IDR frames, -1 indicates codec chosen
    encoder->params.rateControlParams.qpP                     = CONFIGURE(qpP, -1); // Initial QP for P frames
    encoder->params.rateControlParams.qpOffsetB               = CONFIGURE(qpOffsetB, 4); // Offset of B frames QP from P frames
    encoder->params.rateControlParams.qpMaxI                  = CONFIGURE(qpMaxI, 51); // Maximum QP for I/IDR frames
    encoder->params.rateControlParams.qpMinI                  = CONFIGURE(qpMinI, 1); // Minimum QP for I/IDR frames
    encoder->params.rateControlParams.qpMaxP                  = CONFIGURE(qpMaxP, 51); // Maximum QP for P frames
    encoder->params.rateControlParams.qpMinP                  = CONFIGURE(qpMinP, 1); // Minimum QP for P frames
    encoder->params.rateControlParams.qpMaxB                  = CONFIGURE(qpMaxB, 51); // Maximum QP for B frames
    encoder->params.rateControlParams.qpMinB                  = CONFIGURE(qpMinB, 1); // Minimum QP for B frames
    encoder->params.rateControlParams.CbQPIndexOffset         = CONFIGURE(CbQPIndexOffset, 0); // Specifies offset to be added to luma QP for addressing QPC values table for Cb components
    encoder->params.rateControlParams.CrQPIndexOffset         = CONFIGURE(CrQPIndexOffset, 0); // Specifies offset to be added to luma QP for addressing QPC values table for Cr components
    encoder->params.rateControlParams.initialBufferLevel      = CONFIGURE(initialBufferLevel, 2000000); // Initial Buffer level for HRD compliance, keep same as HRDBufferSize for best quality
    encoder->params.rateControlParams.HRDBufferSize           = CONFIGURE(HRDBufferSize, 2000000); // Hypothetical Reference Decoder Buffer Size in bits - 2*bitrate for VBR
    encoder->params.rateControlParams.enablePRC               = CONFIGURE(enablePRC, 1); // Control Flag to enable MB level Perceptual Rate Control, 0 => Disable, 1 => Enable
    encoder->params.rateControlParams.frameSkipAfterSceneChange = CONFIGURE(frameSkipAfterSceneChange, 1); // 0=> no forced skip after scenechange, 1 => force skip frame after coding scene change frame.
    encoder->params.rateControlParams.scalingMatrixPreset      = CONFIGURE(scalingMatrixPreset, 0); // Scaling Matrix selection of encoder, 0: no scaling matrices, 1: normal, 2: noisy, 3 : standard default, 4 : user defined

    // Inter coding

    encoder->params.interCodingParams.interCodingPreset    = CONFIGURE(interCodingPreset, IH264HP_INTERCODING_DEFAULT); // Inter coding mode preset, 0 => deafult values, 1 => user defined
    encoder->params.interCodingParams.MvRangeVerP          = CONFIGURE(MvRangeVerP, 32) ; // Vertical Motion Vector Range for P frames in integer pixels, Range supported is 16 to 496
    encoder->params.interCodingParams.MvRangeHorP          = CONFIGURE(MvRangeHorP, 144); // Horizontal Motion Vector Range for P frames in integer pixels, Range supported is 16 to 496
    encoder->params.interCodingParams.MvRangeVerB          = CONFIGURE(MvRangeVerB, 32) ; // Vertical Motion Vector Range for B frames in integer pixels, Range supported is 16 to 496
    encoder->params.interCodingParams.MvRangeHorB          = CONFIGURE(MvRangeHorB, 144); // Horizontal Motion Vector Range for B frames in integer pixels, Range supported is 16 to 496
    encoder->params.interCodingParams.maxMVperMB           = CONFIGURE(maxMVperMB, 4); // Maximum MV per MB (Values of 1 & 4 are valid)

    // Intra coding

    encoder->params.intraCodingParams.intraCodingPreset             = CONFIGURE(intraCodingPreset, IH264HP_INTRACODING_DEFAULT); // Intra coding mode preset, 0 => deafult values, 1 => user defined
    encoder->params.intraCodingParams.enableIntraPartition          = CONFIGURE(enableIntraPartition, 4); // 0 => INTRA_PARTITION_NONE , 1 => INTRA_PARTITION_ISLICES, 2 => INTRA_PARTITION_IPSLICES, 3 => INTRA_PARTITION_IBSLICES, 4 => INTRA_PARTITION_IPBSLICES.
    encoder->params.intraCodingParams.intraRefreshMethod            = CONFIGURE(intraRefreshMethod, 0); // IH264_INTRAREFRESH_NONE  = 0, IH264_INTRAREFRESH_CYCLIC_MBS = 1
    encoder->params.intraCodingParams.intraRefreshRate              = CONFIGURE(intraRefreshRate, 0); // Rate at which intra MB Refresh is done. e.g. 10 indicates every 10th MB will coded as Intra in inter pictures.
    encoder->params.intraCodingParams.constrainedIntraPredEnable    = CONFIGURE(constrainedIntraPredEnable, 0); // Controls the intra macroblock coding in inter slices

    // Slice coding

    encoder->params.sliceCodingParams.sliceCodingPreset            = CONFIGURE(sliceCodingPreset, IH264HP_SLICECODING_DEFAULT); // Preset value for slice coding mode, 0 => deafult values, 1 => user defined
    encoder->params.sliceCodingParams.streamFormat                 = CONFIGURE(streamFormat, 0); // Type of bitstream to be encoded. Only value 0 is supprted (Byte stream format)
    encoder->params.sliceCodingParams.sliceMode                    = CONFIGURE(sliceMode, 0); // Type of slice coding, 0 => slice coding mode is frame based, 1 => Slices are controlled based upon number of Macroblocks
    encoder->params.sliceCodingParams.sliceUnitSize                = CONFIGURE(sliceUnitSize, 1200); // Number of macroblocks per slice

    // Loop filter

    encoder->params.loopFilterParams.loopfilterPreset               = CONFIGURE(loopfilterPreset, IH264HP_LOOPFILTER_DEFAULT); // Preset value for loop filter operation parameters, 0 => default values, 1 => user defined
    encoder->params.loopFilterParams.loopfilterDisableIDC           = CONFIGURE(loopfilterDisableIDC, 2); // Disable H.264 loop filter, (0=Filter, 1= NoFilter, 2 = No loop filter across slices)
    encoder->params.loopFilterParams.filterOffsetA                  = CONFIGURE(filterOffsetA, 0); // Alpha offset for loop filter
    encoder->params.loopFilterParams.filterOffsetB                  = CONFIGURE(filterOffsetB, 0); // Beta offset for loop filter

    // VUI coding

    encoder->params.vuiCodingParams.vuiCodingPreset                = CONFIGURE(vuiCodingPreset, IH264HP_VUICODING_DEFAULT); // Preset value for VUI Control Params, 0 => deafult values, 1 => user defined
    encoder->params.vuiCodingParams.aspectRatioInfoPresentFlag     = CONFIGURE(aspectRatioInfoPresentFlag, 1); // Controls the insertion of aspect ratio information in VUI part of bit-stream
    encoder->params.vuiCodingParams.aspectRatioIdc                 = CONFIGURE(aspectRatioIdc, IH264HP_ASPECTRATIO_EXTENDED); // Aspect ratio ID. See IH264HPVENC_AspectRatioIdc. IH264HP_ASPECTRATIO_EXTENDED = 255 for dynamic aspect ratio.
    encoder->params.vuiCodingParams.videoSignalTypePresentFlag     = CONFIGURE(videoSignalTypePresentFlag, 0); // controls the insertion of video signal type in VUI part of bit-stream
    encoder->params.vuiCodingParams.videoFormat                    = CONFIGURE(videoFormat, 2); // Video signal type
    encoder->params.vuiCodingParams.videoFullRangeFlag             = CONFIGURE(videoFullRangeFlag, 0); // Flag to specify Range of the pixels
    encoder->params.vuiCodingParams.colourDescriptionPresentFlag   = CONFIGURE(colourDescriptionPresentFlag, 1); // Specifies whether colour_primaries, transfer_characteristics and matrix_coefficients are present or not.
    encoder->params.vuiCodingParams.colourPrimaries                = CONFIGURE(colourPrimaries, 5); // chromaticity coordinates of the source primaries(Table E-3 in standard)
    encoder->params.vuiCodingParams.transferCharacteristics        = CONFIGURE(transferCharacteristics, 5); // opto-electronic transfer characteristic of the source picture(Table E-4 in standard)
    encoder->params.vuiCodingParams.matrixCoefficients             = CONFIGURE(matrixCoefficients, 5); // matrix coefficients used in deriving Y,Cb,Cr signals from the G,B,R primaries(Table E-5 in standard)
    encoder->params.vuiCodingParams.timingInfoPresentFlag          = CONFIGURE(timingInfoPresentFlag, 1); // Controls the insertion of timing info related parameters in VUI part of bit-stream

    // Dynamic parametes

    encoder->dynamicParams.videnc2DynamicParams.forceFrame = IVIDEO_NA_FRAME; // -1: IVIDEO_NA_FRAME, 3: IVIDEO_IDR_FRAME
    encoder->dynamicParams.videnc2DynamicParams.generateHeader = XDM_ENCODE_AU; // 0: Encode entire access unit including headers, 1: Encode only header
    encoder->dynamicParams.videnc2DynamicParams.ignoreOutbufSizeFlag = XDAS_TRUE; // Non configurable
    encoder->dynamicParams.videnc2DynamicParams.inputWidth  = encoderConfiguration->outputWidth;
    encoder->dynamicParams.videnc2DynamicParams.inputHeight = encoderConfiguration->outputHeight;
    encoder->dynamicParams.videnc2DynamicParams.interFrameInterval = encoder->params.videnc2Params.maxInterFrameInterval; // Not configurable
    encoder->dynamicParams.videnc2DynamicParams.intraFrameInterval = CONFIGURE(intraFrameInterval, 1); // The number of frames between two I frames. 0: IPPPP..., 1: IIII..., 2: IPIPIPIPI, 3: IPPIPPIPPI or IPBIPBIPBI, etc.
    encoder->dynamicParams.videnc2DynamicParams.mvAccuracy = CONFIGURE(mvAccuracy, IVIDENC2_MOTIONVECTOR_QUARTERPEL); // Motion vectors accuracy. 0: integer pel., 2: quarter pel.
    encoder->dynamicParams.videnc2DynamicParams.putDataFxn = NULL;
    encoder->dynamicParams.videnc2DynamicParams.putDataHandle = 0;
    encoder->dynamicParams.videnc2DynamicParams.getDataFxn = NULL;
    encoder->dynamicParams.videnc2DynamicParams.getDataHandle = 0;
    encoder->dynamicParams.videnc2DynamicParams.getBufferFxn = NULL;
    encoder->dynamicParams.videnc2DynamicParams.getBufferHandle = 0;
    encoder->dynamicParams.videnc2DynamicParams.refFrameRate = encoderConfiguration->outputFrameRate * 1000;
    encoder->dynamicParams.videnc2DynamicParams.targetFrameRate = encoderConfiguration->outputFrameRate * 1000;
    encoder->dynamicParams.videnc2DynamicParams.sampleAspectRatioWidth = 1;
    encoder->dynamicParams.videnc2DynamicParams.sampleAspectRatioHeight = 1;
    encoder->dynamicParams.videnc2DynamicParams.targetBitRate = encoder->params.videnc2Params.maxBitRate;

    encoder->dynamicParams.transformBlockSize = encoder->params.transformBlockSize;

    memcpy(&encoder->dynamicParams.rateControlParams, &encoder->params.rateControlParams, sizeof (IH264HPVENC_RateControlParams));
    memcpy(&encoder->dynamicParams.interCodingParams, &encoder->params.interCodingParams, sizeof (IH264HPVENC_InterCodingParams));
    memcpy(&encoder->dynamicParams.intraCodingParams, &encoder->params.intraCodingParams, sizeof (IH264HPVENC_IntraCodingParams));
    memcpy(&encoder->dynamicParams.sliceCodingParams, &encoder->params.sliceCodingParams, sizeof (IH264HPVENC_SliceCodingParams));
}

static void H264HighProfileEncoder_delete(Encoder2 *encoder)
{
    int result;
    H264HighProfileEncoder *h264Encoder = (H264HighProfileEncoder*) encoder;

    if (h264Encoder != NULL) {
        if (h264Encoder->isResourcesAssigned) {
            encoder->activate(encoder);
            result = RMAN_deactivateAllResources((IALG_Handle) encoder->handle, &H264HPVENC_TI_IRES, 0);
            encoder->deactivate(encoder);
            if (result != IRES_OK) {
                printf("WARNING: Deactivate resources failed (%d)\n", result);
            }

            encoder->activate(encoder);
            result = RMAN_freeResources((IALG_Handle) encoder->handle, &H264HPVENC_TI_IRES, 0);
            encoder->deactivate(encoder);
            if (result != IRES_OK) {
                printf("WARNING: Free resources failed (%d)\n", result);
            }
        }

        if (h264Encoder->encoder.handle != NULL) {
            MulticoreApi_setActiveAlgorithm(encoder->algorithmIndex);
            ResultCode result = Algorithm_delete((IALG_Handle) encoder->handle);
            MulticoreApi_setActiveAlgorithm(-1);
            if (result != ResultCode_Ok) {
                puts("Failed to delete encoder");
            }
        }

        free(h264Encoder);
    }
}

Encoder2* H264HighProfileEncoder_create(const DspEncoderConfiguration *encoderConfiguration, int encoderIndex)
{
    int result;
    IVIDENC2_Fxns* encoderFunctions;
    H264HighProfileEncoder *h264Encoder = malloc(sizeof(H264HighProfileEncoder));
    int numberOfEncoderCores = Configuration_calculateNumberOfCores(encoderConfiguration->coreIds);

    if (h264Encoder == NULL) {
        puts("Failed to allocate memory for encoder");
        goto CREATE_ERROR;
    }

    memset(h264Encoder, 0, sizeof(H264HighProfileEncoder));

    h264Encoder->encoder.inArguments = &h264Encoder->inArguments.videnc2InArgs;
    h264Encoder->encoder.outArguments = &h264Encoder->outArguments.videnc2OutArgs;
    h264Encoder->encoder.status = &h264Encoder->status.videnc2Status;
    h264Encoder->encoder.params = &h264Encoder->params.videnc2Params;
    h264Encoder->encoder.dynamicParams = &h264Encoder->dynamicParams.videnc2DynamicParams;
    h264Encoder->encoder.delete = H264HighProfileEncoder_delete;

    h264Encoder->status.videnc2Status.size = sizeof(h264Encoder->status);
    h264Encoder->inArguments.videnc2InArgs.size = sizeof(h264Encoder->inArguments);
    h264Encoder->outArguments.videnc2OutArgs.size = sizeof(h264Encoder->outArguments);

    H264HighProfileEncoder_configure(h264Encoder, encoderConfiguration);

    if (H264HighProfileEncoder_initializeMulticoreApi(h264Encoder, encoderConfiguration, encoderIndex) != ResultCode_Ok) {
        goto CREATE_ERROR;
    }

    MulticoreApi_setActiveAlgorithm(h264Encoder->encoder.algorithmIndex);

    h264Encoder->encoder.handle = (IVIDENC2_Handle) Algorithm_create(
            &H264HPVENC_TI_IH264HPVENC.ividenc.ialg,
            (IALG_Params*) &h264Encoder->params,
            h264Encoder->encoder.memTabs);

    MulticoreApi_setActiveAlgorithm(-1);

    if (h264Encoder->encoder.handle == NULL) {
        puts("Failed to create encoder");
        goto CREATE_ERROR;
    }

    h264Encoder->encoder.functions = h264Encoder->encoder.handle->fxns;
    encoderFunctions = h264Encoder->encoder.functions;

    // Configure encoder resources

    h264Encoder->encoder.activate(&h264Encoder->encoder);
    result = RMAN_assignResources((IALG_Handle) h264Encoder->encoder.handle, &H264HPVENC_TI_IRES, 0);

    if (result != IRES_OK) {
        h264Encoder->encoder.deactivate(&h264Encoder->encoder);
        printf("Assign resources failed (%d)\n", result);
        goto CREATE_ERROR;
    }

    h264Encoder->isResourcesAssigned = 1;

    result = RMAN_activateAllResources((IALG_Handle) h264Encoder->encoder.handle, &H264HPVENC_TI_IRES, 0);
    h264Encoder->encoder.deactivate(&h264Encoder->encoder);

    //if (result != IRES_OK) {
        // TODO: WARNING: fix this
    //}

    // Configure encoder

    h264Encoder->encoder.activate(&h264Encoder->encoder);
    result = encoderFunctions->control(h264Encoder->encoder.handle, XDM_RESET, &h264Encoder->dynamicParams.videnc2DynamicParams, &h264Encoder->status.videnc2Status);
    h264Encoder->encoder.deactivate(&h264Encoder->encoder);

    if (result != IVIDENC_EOK) {
        printf("XDM_RESET failed (%d) with extended error (%d)\n", result, h264Encoder->status.videnc2Status.extendedError);
        goto CREATE_ERROR;
    }

    // Set params

    h264Encoder->encoder.activate(&h264Encoder->encoder);
    result = encoderFunctions->control(h264Encoder->encoder.handle, XDM_SETPARAMS, &h264Encoder->dynamicParams.videnc2DynamicParams, &h264Encoder->status.videnc2Status);
    h264Encoder->encoder.deactivate(&h264Encoder->encoder);

    if (result != IVIDENC_EOK) {
        printf("XDM_SETPARAMS failed (%d) with extended error (%d)\n", result, h264Encoder->status.videnc2Status.extendedError);
        goto CREATE_ERROR;
    }

    MulticoreApi_setActiveAlgorithm(h264Encoder->encoder.algorithmIndex);
    h264Encoder->encoder.shmmap_sync(h264Encoder->encoder.encoderTaskIndex, IVIDMC_SHMEM_KEY_CLEAN, NULL);
    Algorithm_synchronize(h264Encoder->encoder.memTabs);
    MulticoreApi_swbarr(DNUM, 0, numberOfEncoderCores);
    MulticoreApi_setActiveAlgorithm(-1);

    return (Encoder2*) h264Encoder;

CREATE_ERROR:

    H264HighProfileEncoder_delete((Encoder2*) h264Encoder);
    return NULL;
}
