/**
 * Multicore API for algorithms (i.e. decoders and encoders).
 */

#include <stdio.h>
#include <string.h>

#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>
#include <ti/sysbios/family/c66/Cache.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>

#include "GlobalGates.h"
#include "DspCommonTypes.h"
#include "Heaps.h"
#include "MulticoreApi.h"

#define MAX_ALGORITHMS        (MAX_ENCODERS_PER_DSP + MAX_DECODERS_PER_DSP)
#define BARRIER_AVAILABLE     (0)
#define BARRIER_NOT_AVAILABLE (1)

// Defines from ividimc.h header. There are several ividimc.h that contain very similar code.
#define IVIDMC_SHMEM_KEY_CLEAN          (0xFFFF)
#define IVIDMC_SHMEMKEY_FIRST           (0)
#define IVIDMC_SHMEM_NUM_KEYS           (16)
#define IVIDMC_NUM_BARRIERS             (16)
#define IVIDMC_SHMEM_ATTRS_DDR_CACHED   (0)
#define IVIDMC_SHMEM_ATTRS_DDR_UNCACHED (1)
#define IVIDMC_SHMEM_ATTRS_SL2          (2)

#define RESERVATION_WAIT_INTERVAL (1000 / Clock_tickPeriod) // 1 millisecond

// Variables shared between cores

#pragma DATA_ALIGN(barrierCounters, L2_CACHE_LINE_SIZE)
#pragma DATA_SECTION(barrierCounters, ".msmc")
static int barrierCounters[MAX_ALGORITHMS][IVIDMC_NUM_BARRIERS];

#pragma DATA_ALIGN(sharedMemoryBases, L2_CACHE_LINE_SIZE)
#pragma DATA_SECTION(sharedMemoryBases, ".msmc")
static void* sharedMemoryBases[MAX_ALGORITHMS][IVIDMC_SHMEM_NUM_KEYS];

#pragma DATA_ALIGN(sharedMemorySizes, L2_CACHE_LINE_SIZE)
#pragma DATA_SECTION(sharedMemorySizes, ".msmc")
static int32_t sharedMemorySizes[MAX_ALGORITHMS][IVIDMC_SHMEM_NUM_KEYS];

#pragma DATA_ALIGN(sharedMemoryHeaps, L2_CACHE_LINE_SIZE)
#pragma DATA_SECTION(sharedMemoryHeaps, ".msmc")
static IHeap_Handle sharedMemoryHeaps[MAX_ALGORITHMS][IVIDMC_SHMEM_NUM_KEYS];

typedef struct {
    int32_t userId;
} Ividmc3KeyLockData;

typedef struct {
    int32_t usersCount;
    int32_t counter;
} Ividmc3KeySwbarrData;

typedef struct {
    void* base;
    IHeap_Handle heap;
    int32_t size;
} Ividmc3KeyShmemData;

typedef union {
    Ividmc3KeyLockData lock;
    Ividmc3KeySwbarrData swbarr;
    Ividmc3KeyShmemData shmem;
} Ividmc3KeyData;

typedef struct Ividmc3Key_s {
    char name[32];
    IVIDMC3_KEY_SPACE_e keySpace;
    struct Ividmc3Key_s* next;
    Ividmc3KeyData data;
} Ividmc3Key;

#pragma DATA_ALIGN(ividmc3KeyChains, L2_CACHE_LINE_SIZE)
#pragma DATA_SECTION(ividmc3KeyChains, ".msmc")
static Ividmc3Key* ividmc3KeyChains[MAX_ALGORITHMS];

// Variables not shared between cores

int MulticoreApi_activeAlgorithm = -1;

void MulticoreApi_swbarrWithYield(int32_t algorithmIndex, uint32_t swbarr_cnt)
{
    GateMP_Handle gate = GlobalGates_getMulticoreApiGate();
    IArg gateKey;
    int* barrierCounter = &barrierCounters[algorithmIndex][0];

    // Wait while the barrier became available
    while (1) {
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
        if (*barrierCounter >= 0) {
            break;
        }
        Task_sleep(1);
    }

    // Enter gate
    gateKey = GateMP_enter(gate);
    Cache_inv(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);

    if (*barrierCounter == swbarr_cnt - 1) { // Last task entered barrier
        *barrierCounter = -swbarr_cnt + 1; // Leave barrier
    } else {
        (*barrierCounter)++;
        Cache_wb(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);
        GateMP_leave(gate, gateKey);

        // Wait for last task
        while (1) {
            Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
            if (*barrierCounter < 0) {
                break;
            }
            Task_sleep(1);
        }

        gateKey = GateMP_enter(gate);
        Cache_inv(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);

        (*barrierCounter)++; // Leave barrier
    }

    Cache_wb(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

void MulticoreApi_swbarr(int32_t coreID, int32_t swbarr_id, uint32_t swbarr_cnt)
{
    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey;
    int* barrierCounter = &barrierCounters[MulticoreApi_activeAlgorithm][swbarr_id];

    // Wait while the barrier became available
    do {
        Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
    } while (*barrierCounter < 0);

    // Enter gate
    gateKey = GateMP_enter(gate);
    Cache_inv(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);

    if (*barrierCounter == swbarr_cnt - 1) { // Last task entered barrier
        *barrierCounter = -swbarr_cnt + 1; // Leave barrier
    } else {
        (*barrierCounter)++;
        Cache_wb(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);
        GateMP_leave(gate, gateKey);

        // Wait for last task
        do {
            Cache_inv(barrierCounter, sizeof(*barrierCounter), Cache_Type_ALL, TRUE);
        } while (*barrierCounter >= 0);

        gateKey = GateMP_enter(gate);
        Cache_inv(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);

        (*barrierCounter)++; // Leave barrier
    }

    Cache_wb(barrierCounters, sizeof(barrierCounters), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);
}

int32_t* MulticoreApi_shmmap(int32_t coreID, int32_t shmem_key, int32_t shmem_size, uint32_t attr, int32_t alignment)
{
    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey = GateMP_enter(gate);

    Cache_inv(sharedMemoryBases, sizeof(sharedMemoryBases), Cache_Type_ALL, TRUE);

    void** sharedMemoryBase = &sharedMemoryBases[MulticoreApi_activeAlgorithm][shmem_key];

    if (*sharedMemoryBase != NULL) {
        GateMP_leave(gate, gateKey);
        return *sharedMemoryBase;
    }

    Error_Block eb;
    Error_init(&eb);

    int rest = shmem_size % 16;

    if (rest != 0) {
        shmem_size = shmem_size + 16 - rest;
    }

    IHeap_Handle heap = (attr == IVIDMC_SHMEM_ATTRS_SL2) ? msmcHeap : ddr3Heap;
    *sharedMemoryBase = Memory_alloc(heap, shmem_size, (alignment <= L2_CACHE_LINE_SIZE) ? L2_CACHE_LINE_SIZE : alignment, &eb);

    Cache_inv(sharedMemorySizes, sizeof(sharedMemorySizes), Cache_Type_ALL, TRUE);
    Cache_inv(sharedMemoryHeaps, sizeof(sharedMemoryHeaps), Cache_Type_ALL, TRUE);

    int32_t *sharedMemorySize = &sharedMemorySizes[MulticoreApi_activeAlgorithm][shmem_key];
    IHeap_Handle* sharedMemoryHeap = &sharedMemoryHeaps[MulticoreApi_activeAlgorithm][shmem_key];

    *sharedMemorySize = shmem_size;
    *sharedMemoryHeap = heap;

    Cache_wb(sharedMemoryBases, sizeof(sharedMemoryBases), Cache_Type_ALL, TRUE);
    Cache_wb(sharedMemorySizes, sizeof(sharedMemorySizes), Cache_Type_ALL, TRUE);
    Cache_wb(sharedMemoryHeaps, sizeof(sharedMemoryHeaps), Cache_Type_ALL, TRUE);

    GateMP_leave(gate, gateKey);
    return *sharedMemoryBase;
}

int32_t MulticoreApi_shmmunmap(int32_t coreID, int32_t shmem_key, int32_t *shmem_base)
{
    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey = GateMP_enter(gate);

    // TODO: optimize to one invalidate call
    Cache_inv(sharedMemoryBases, sizeof(sharedMemoryBases), Cache_Type_ALL, TRUE);
    Cache_inv(sharedMemorySizes, sizeof(sharedMemorySizes), Cache_Type_ALL, TRUE);
    Cache_inv(sharedMemoryHeaps, sizeof(sharedMemoryHeaps), Cache_Type_ALL, TRUE);

    void** sharedMemoryBase = &sharedMemoryBases[MulticoreApi_activeAlgorithm][shmem_key];
    int32_t* sharedMemorySize = &sharedMemorySizes[MulticoreApi_activeAlgorithm][shmem_key];
    IHeap_Handle* sharedMemoryHeap = &sharedMemoryHeaps[MulticoreApi_activeAlgorithm][shmem_key];

    Memory_free(*sharedMemoryHeap, *sharedMemoryBase, *sharedMemorySize);

    *sharedMemoryBase = NULL;
    Cache_wb(sharedMemoryBases, sizeof(sharedMemoryBases), Cache_Type_ALL, TRUE);

    GateMP_leave(gate, gateKey);
    return 0;
}

int32_t MulticoreApi_shmmap_sync(int32_t coreID, int32_t shmem_key, int32_t *shmem_base)
{
    if (shmem_key == IVIDMC_SHMEM_KEY_CLEAN) {
        int key;
        for (key = IVIDMC_SHMEMKEY_FIRST; key < IVIDMC_SHMEM_NUM_KEYS; key++) {
            void *base = sharedMemoryBases[MulticoreApi_activeAlgorithm][key];
            if (base != NULL) {
                Cache_wbInv(base, sharedMemorySizes[MulticoreApi_activeAlgorithm][key], Cache_Type_ALL, TRUE);
            }
        }
    } else  {
        Cache_wbInv(shmem_base, sharedMemorySizes[MulticoreApi_activeAlgorithm][shmem_key], Cache_Type_ALL, TRUE);
    }

    return 0;
}

void MulticoreApi_swbarrDummy(int32_t coreID, int32_t swbarr_id, uint32_t swbarr_cnt)
{
}

void MulticoreApi_swbarrWithYieldDummy(int32_t algorithmIndex, uint32_t swbarr_cnt)
{
}

int32_t* MulticoreApi_shmmapDummy(int32_t coreID, int32_t shmem_key, int32_t shmem_size, uint32_t attr, int32_t alignment)
{
    return NULL;
}

int32_t MulticoreApi_shmmunmapDummy(int32_t coreID, int32_t shmem_key, int32_t *shmem_base)
{
    return 0;
}

int32_t MulticoreApi_shmmap_syncDummy(int32_t coreID, int32_t shmem_key, int32_t *shmem_base)
{
    return 0;
}

static Ividmc3Key* findIvidmc3Key(const uint8_t* name, IVIDMC3_KEY_SPACE_e keySpace)
{
    Ividmc3Key* currentKey = ividmc3KeyChains[MulticoreApi_activeAlgorithm];

    while (currentKey != NULL) {
        Cache_inv(currentKey, sizeof(*currentKey), Cache_Type_ALL, TRUE);

        if ((currentKey->keySpace == keySpace) && (strcmp(&currentKey->name[0], (const char*) name) == 0)) {
            break;
        }

        currentKey = currentKey->next;
    }

    return currentKey;
}

static void addIvidmc3Key(Ividmc3Key* key)
{
    Ividmc3Key* currentKey = ividmc3KeyChains[MulticoreApi_activeAlgorithm];

    if (currentKey == NULL) {
        ividmc3KeyChains[MulticoreApi_activeAlgorithm] = key;
        Cache_wb(ividmc3KeyChains, sizeof(ividmc3KeyChains), Cache_Type_ALL, TRUE);
    } else {
        while (1) {
            Cache_inv(currentKey, sizeof(*currentKey), Cache_Type_ALL, TRUE);

            if (currentKey->next == NULL) {
                break;
            }

            currentKey = currentKey->next;
        }

        currentKey->next = key;
        Cache_wb(currentKey, sizeof(*currentKey), Cache_Type_ALL, TRUE);
    }
}

static Ividmc3Key* createIvidmc3Key(const uint8_t *name, IVIDMC3_KEY_SPACE_e keySpace, Ividmc3KeyData* data)
{
    Error_Block eb;
    Error_init(&eb);

    Ividmc3Key* key = Memory_alloc(msmcHeap, sizeof(Ividmc3Key), L2_CACHE_LINE_SIZE, &eb);

    if (key != NULL) {
        strcpy(&key->name[0], (const char*) name);
        key->keySpace = keySpace;
        key->data = *data;
        key->next = NULL;
        Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
    }

    return key;
}

static Ividmc3Key* createLock(const uint8_t *name, int32_t user_id, IVIDMC3_KEY_SPACE_e keySpace, int32_t num_users, int32_t *user_ids, IVIDMC3_KeyCfg_t *cfg)
{
    Ividmc3KeyData data;
    data.lock.userId = -1;
    return createIvidmc3Key(name, keySpace, &data);
}

static Ividmc3Key* createBarrier(const uint8_t *name, int32_t user_id, IVIDMC3_KEY_SPACE_e keySpace, int32_t num_users, int32_t *user_ids, IVIDMC3_KeyCfg_t *cfg)
{
    Ividmc3KeyData data;
    data.swbarr.counter = 0;
    data.swbarr.usersCount = num_users;
    return createIvidmc3Key(name, keySpace, &data);
}

static Ividmc3Key* createShmem(const uint8_t *name, int32_t user_id, IVIDMC3_KEY_SPACE_e keySpace, int32_t num_users, int32_t *user_ids, IVIDMC3_KeyCfg_t *cfg)
{
    Ividmc3KeyData data;

    switch (cfg->u.shm.attr) {
    case IVIDMC3_SHMEM_ATTRS_DDR_CACHED:
        data.shmem.heap = ddr3Heap;
        break;

    case IVIDMC3_SHMEM_ATTRS_DDR_UNCACHED:
        data.shmem.heap = ddr3UncachedHeap;
        break;

    case IVIDMC3_SHMEM_ATTRS_SL2:
        data.shmem.heap = msmcHeap;
        break;

    default:
        return NULL;
    }

    Error_Block eb;
    Error_init(&eb);

    data.shmem.size = cfg->u.shm.size;
    data.shmem.base = Memory_alloc(data.shmem.heap, data.shmem.size, cfg->u.shm.alignment, &eb);

    if (data.shmem.base == NULL) {
        puts("Error allocating memory");
        fflush(stdout);
        return NULL;
    }

    return createIvidmc3Key(name, keySpace, &data);
}

void* MulticoreApi3_keyCreate(uint8_t *name, int32_t user_id, IVIDMC3_KEY_SPACE_e key_space, int32_t num_users, int32_t *user_ids, IVIDMC3_KeyCfg_t *cfg)
{
    //printf("keyCreate name='%s' userid=%d keyspace=%d numusers=%d\n", name, user_id, key_space, num_users);
    //printf("alignment=%d size=%d attr=%d\n", cfg->u.shm.alignment, cfg->u.shm.size, cfg->u.shm.attr);
    //fflush(stdout);

    if (strlen((char*) name) >= sizeof(((Ividmc3Key*)NULL)->name)) {
        return NULL;
    }

    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(ividmc3KeyChains, sizeof(ividmc3KeyChains), Cache_Type_ALL, TRUE);

    Ividmc3Key* key = findIvidmc3Key(name, key_space);

    if (key == NULL) {
        switch (key_space) {
        case IVIDMC3_LOCK_SPACE:
            key = /*((num_users > 1) ? */createLock(name, user_id, key_space, num_users, user_ids, cfg)/* : (Ividmc3Key*) 1)*/;
            break;

        case IVIDMC3_SWBARR_SPACE:
            key = /*((num_users > 1) ? */createBarrier(name, user_id, key_space, num_users, user_ids, cfg)/* : (Ividmc3Key*) 1)*/;
            break;

        case IVIDMC3_SHMEM_SPACE:
            key = createShmem(name, user_id, key_space, num_users, user_ids, cfg);
            break;

        default:
            key = NULL;
            break;
        }

        if ((key != NULL)/* && ((num_users > 1) || (key_space == IVIDMC3_SHMEM_SPACE))*/) {
            addIvidmc3Key(key);
        }
    }

    GateMP_leave(gate, gateKey);

    //printf("result=%p\n", key);
    //fflush(stdout);
    return key;
}

/*int32_t MulticoreApi3_lockAcquire(int32_t user_id, void* lockHandle)
{
    printf("lockAcquire userid=%d handle=%p\n", user_id, lockHandle);

    if (lockHandle == NULL) {
        return IVIDMC3_LOCK_ERROR_HANDLE;
    }

    Ividmc3Key* key = (Ividmc3Key*) lockHandle;

    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey = GateMP_enter(gate);
    Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

    int32_t result;

    if (key->data.lock.userId == -1) { // Lock is released
        key->data.lock.userId = user_id;
        Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
        result = 1;
    } else {
        result = 0;
    }

    GateMP_leave(gate, gateKey);
    return result;
}*/

int32_t MulticoreApi3_lockAcquire(int32_t user_id, void* lockHandle)
{
    //printf("lockAcquire userid=%d handle=%p\n", user_id, lockHandle);
    //fflush(stdout);

    if (lockHandle == NULL) {
        return IVIDMC3_LOCK_ERROR_HANDLE;
    }

    Ividmc3Key* key = (Ividmc3Key*) lockHandle;

    while (1) {
        GateMP_Handle gate = GlobalGates_getBarrierGate();
        IArg gateKey = GateMP_enter(gate);

        Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

        if (key->data.lock.userId == -1) { // Lock is released
            key->data.lock.userId = user_id;
            Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
            GateMP_leave(gate, gateKey);
            //printf("result=%d\n", IVIDMC3_LOCK_ERROR_NONE);
            //fflush(stdout);
            return IVIDMC3_LOCK_ERROR_NONE;
        } /*else if (key->data.lock.userId == user_id) { // Already acquired by the same user
            GateMP_leave(gate, gateKey);
            printf("result=%d\n", IVIDMC3_LOCK_ERROR_DBL_LOCK);
            fflush(stdout);
            return IVIDMC3_LOCK_ERROR_DBL_LOCK;
        }*/

        GateMP_leave(gate, gateKey);

        // Wait until lock is released
        do {
            Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);
        } while (key->data.lock.userId == -1);
    }
}

int32_t MulticoreApi3_lockRelease(int32_t user_id, void* lockHandle)
{
    //printf("lockRelease userid=%d handle=%p\n", user_id, lockHandle);
    //fflush(stdout);

    Ividmc3Key* key = (Ividmc3Key*) lockHandle;
    int32_t result;

    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey = GateMP_enter(gate);

    Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

    if (key->data.lock.userId == user_id) {
        result = IVIDMC3_LOCK_ERROR_NONE;
        key->data.lock.userId = -1;
        Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
    } else if (key->data.lock.userId == -1) {
        result = IVIDMC3_LOCK_ERROR_DBL_UNLOCK;
    } else {
        result = IVIDMC3_LOCK_ERROR_USER_ID;
    }

    GateMP_leave(gate, gateKey);

    //printf("result=%d\n", result);
    //fflush(stdout);
    return result;
}

int32_t MulticoreApi3_lockCheck(int32_t user_id, void* lockHandle)
{
    //printf("lockCheck userid=%d handle=%p\n", user_id, lockHandle);
    //fflush(stdout);

    Ividmc3Key* key = (Ividmc3Key*) lockHandle;
    int32_t result;

    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey = GateMP_enter(gate);

    Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

    result = (key->data.lock.userId == -1) ? IVIDMC3_LOCK_ERROR_NONE : IVIDMC3_LOCK_IN_USE;

    GateMP_leave(gate, gateKey);

    //printf("result=%d\n", result);
    //fflush(stdout);
    return result;
}

int32_t MulticoreApi3_barrWait(int32_t user_id, void* barrHandle)
{
    //printf("barrWait userid=%d handle=%p\n", user_id, barrHandle);
    //fflush(stdout);

    Ividmc3Key* key = (Ividmc3Key*) barrHandle;

    GateMP_Handle gate = GlobalGates_getBarrierGate();
    IArg gateKey;

    // Wait while the barrier became available
    do {
        Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);
    } while (key->data.swbarr.counter < 0);

    // Enter gate
    gateKey = GateMP_enter(gate);
    Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

    int32_t usersCount = key->data.swbarr.usersCount;

    if (key->data.swbarr.counter == usersCount - 1) { // Last task entered barrier
        key->data.swbarr.counter = -usersCount + 1; // Leave barrier
    } else {
        key->data.swbarr.counter++;
        Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
        GateMP_leave(gate, gateKey);

        // Wait for last task
        do {
            Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);
        } while (key->data.swbarr.counter >= 0);

        gateKey = GateMP_enter(gate);
        Cache_inv(key, sizeof(*key), Cache_Type_ALL, TRUE);

        key->data.swbarr.counter++; // Leave barrier
    }

    Cache_wb(key, sizeof(*key), Cache_Type_ALL, TRUE);
    GateMP_leave(gate, gateKey);

    return IVIDMC3_BARR_ERROR_NONE;
}

int32_t* MulticoreApi3_shmMap(int32_t user_id, void* shmemHandle)
{
    //printf("shmMap userid=%d handle=%p\n", user_id, shmemHandle);
    //printf("result=%p\n", ((Ividmc3Key*) shmemHandle)->data.shmem.base);
    //fflush(stdout);
    return ((Ividmc3Key*) shmemHandle)->data.shmem.base;
}

int32_t MulticoreApi3_shmSync(int32_t user_id, void* shmemHandle, int32_t* shmem_base, int32_t shmem_size, uint32_t shmem_sync_attribs)
{
    //printf("shmSync userid=%d handle=%p\n", user_id, shmemHandle);
    //fflush(stdout);
    if (shmem_sync_attribs & IVIDMC3_SYNC_ATTRIBS_LOC_WRITEBACK) {
        Cache_wbInv(shmem_base, shmem_size, Cache_Type_ALL, TRUE);
    } else if (shmem_sync_attribs & IVIDMC3_SYNC_ATTRIBS_LOC_INVALIDATE) {
        Cache_inv(shmem_base, shmem_size, Cache_Type_ALL, TRUE);
    }

    return 0;
}

int32_t MulticoreApi3_shmSyncWait(int32_t user_id, void* shmemHandle, int32_t* shmem_transid)
{
    //printf("shmSyncWait userid=%d handle=%p\n", user_id, shmemHandle);
    //fflush(stdout);
    return 0;
}
