/**
 * Contains main() function and SYS/BIOS start routines.
 */

#include <stdio.h>

#include <xdc/runtime/Log.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/IHeap.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/family/c66/Cache.h>
#include <ti/sysbios/heaps/HeapMem.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sdo/fc/rman/rman.h>
#include <ti/csl/csl_bootcfgAux.h>

#include <ti/ipc/Ipc.h>
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/HeapMemMP.h>
#include <ti/ipc/MessageQ.h>

#include "EdmaConfig.h"
#include "Heaps.h"
#include "Configuration.h"
#include "GlobalGates.h"
#include "Common.h"

extern cregister volatile unsigned int DNUM;

#define DEVICE_REG32_W(x,y) *(volatile uint32_t *)(x)=(y)
#define IPCGR(x)            (0x02620240 + x*4)

extern void encodeMasterTask(UArg arg0, UArg arg1);
extern void decodeMasterTask(UArg arg0, UArg arg1);
extern void decodeSlaveTask(UArg arg0, UArg arg1);

static ResultCode initializeRman()
{
    int edmaInstanceId = (DNUM & 1) + 1; // Balancing of EDMA instances (1 and 2) between cores
    int coreId = DNUM;
    int queueId = 0;

    printf("Using EDMA3 instance %d\n", edmaInstanceId);

    // Specify EDMA instance ID and queue ID
    EDMA3_PARAMS.regionConfig = &C6608_config[edmaInstanceId][coreId];
    ti_sdo_fc_edma3_EDMA3_physicalId = edmaInstanceId;
    ti_sdo_fc_edma3_Settings_defaultQ = queueId;

    int result = RMAN_init();

    if (result != IRES_OK) {
        printf("RMAN initialization failed (%d)\n", result);
        return ResultCode_Error;
    }

    return ResultCode_Ok;
}

static void initializeCaching()
{
    Cache_Size size;
    size.l1pSize  = Cache_L1Size_32K; // L1P cache size
    size.l1dSize  = Cache_L1Size_32K; // L1D cache size
    size.l2Size   = Cache_L2Size_64K; // L2  cache size
    Cache_setSize(&size);

    UInt32 mar = Cache_getMar((Ptr) ddr3HeapRegionStartAddress) | ti_sysbios_family_c66_Cache_PC | ti_sysbios_family_c66_Cache_PFX;
    Cache_setMar((Ptr) ddr3HeapRegionStartAddress, ddr3HeapRegionSize, Cache_Mar_ENABLE | mar);

    const UInt32 disablePrefetch = (~ti_sysbios_family_c66_Cache_PFX) & (0xFFFFFFFF);
    const UInt32 disableCache = (~ti_sysbios_family_c66_Cache_PC) & (0xFFFFFFFF);
    mar = Cache_getMar((Ptr) ddr3UncachedHeapRegionStartAddress) & disableCache & disablePrefetch;
    Cache_setMar((Ptr) ddr3UncachedHeapRegionStartAddress, ddr3UncachedHeapRegionSize, Cache_Mar_DISABLE | mar);

    Cache_wbInvAll();
}

static ResultCode createSharedHeap(IHeap_Handle* heapHandle, int heapSize, char* heapName, int regionId, int heapId)
{
    if (DNUM == 0) {
        // Create heap

        HeapMemMP_Params params;
        HeapMemMP_Params_init(&params);

        params.regionId = regionId;
        params.name = heapName;
        params.sharedBufSize = heapSize - 0x100; // Minus heap creation overhead

        *heapHandle = (IHeap_Handle) HeapMemMP_create(&params);
        if (*heapHandle == NULL) {
            puts("Failed to create shared heap");
            return ResultCode_Error;
        }
    } else {
        // Open heap

        while (1) {
            if (HeapMemMP_open(heapName, (HeapMemMP_Handle*) heapHandle) == HeapMemMP_S_SUCCESS) {
                break;
            }
        }
    }

    if (MessageQ_registerHeap(*heapHandle, heapId) != MessageQ_S_SUCCESS) {
        puts("Failed to create shared heap");
        return ResultCode_Error;
    }

    return ResultCode_Ok;
}

static ResultCode processConfiguration()
{
    int i;
    Error_Block eb;
    const DspConfiguration* configuration = Configuration_get();

    Error_init(&eb);

    // Initialize core statistics data

    int decodersCount = 0;
    int encodersCount = 0;

    for (i = 0; i < configuration->decodersCount; i++) {
        int coreIndex;
        for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
            int coreId = configuration->decoderConfigurations[i].coreIds[coreIndex];

            if (coreId == -1) {
                break;
            } else if (coreId != DNUM) {
                continue;
            }

            decodersCount++;
        }
    }

    for (i = 0; i < configuration->encodersCount; i++) {
        int coreIndex;
        for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
            int coreId = configuration->encoderConfigurations[i].coreIds[coreIndex];

            if (coreId == -1) {
                break;
            } else if (coreId != DNUM) {
                continue;
            }

            encodersCount++;
        }
    }

    for (i = 0; i < configuration->decodersCount; i++) {
        int coreIndex;
        for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
            int coreId = configuration->decoderConfigurations[i].coreIds[coreIndex];

            if (coreId == -1) {
                break;
            } else if (coreId != DNUM) {
                continue;
            }

            printf("Creating decoder task for decoder #%d\n", i);
            fflush(stdout);

            Task_Params taskParams;
            Task_Params_init(&taskParams);
            taskParams.arg0 = i;
            taskParams.arg1 = 0;
            Task_FuncPtr taskFunction;

            if (coreIndex == 0) { // Master decoder
                taskFunction = decodeMasterTask;
                taskParams.stackSize = 0x4000;
            } else { // Slave decoder
                taskFunction = decodeSlaveTask;
                taskParams.stackSize = 0x4000;
            }

            Task_Handle taskHandle = Task_create(taskFunction, &taskParams, &eb);

            if (taskHandle == NULL) {
                puts("Failed to create task");
                return ResultCode_Error;
            }
        }
    }

    for (i = 0; i < configuration->encodersCount; i++) {
        int coreIndex;
        for (coreIndex = 0; coreIndex < CORES_PER_DSP; coreIndex++) {
            int coreId = configuration->encoderConfigurations[i].coreIds[coreIndex];

            if (coreId == -1) {
                break;
            } else if (coreId != DNUM) {
                continue;
            }

            printf("Creating encoder task for encoder #%d\n", i);
            fflush(stdout);

            Task_Params taskParams;
            Task_Params_init(&taskParams);
            taskParams.arg0 = i;
            taskParams.arg1 = 0;
            Task_FuncPtr taskFunction;

            if (coreIndex == 0) { // Master encoder
                taskFunction = encodeMasterTask;
                taskParams.stackSize = 0x4000;
            } else { // Slave encoder
                puts("Slave encoder not supported");
                return ResultCode_Error;
            }

            Task_Handle taskHandle = Task_create(taskFunction, &taskParams, &eb);

            if (taskHandle == NULL) {
                puts("Failed to create task");
                return ResultCode_Error;
            }
        }
    }

    return ResultCode_Ok;
}

static ResultCode attachIpc()
{
    int i;

    for (i = 0; i < MultiProc_getNumProcessors(); i++) {
        if (i == MultiProc_self())  {
            continue;
        }

        printf("IPC: attaching to core#%d\n", i);
        fflush(stdout);

        int result;
        do {
            result = Ipc_attach(i);
        } while (result == Ipc_E_NOTREADY);

        if (result != Ipc_S_SUCCESS) {
            return ResultCode_Error;
        }

        puts("Attached");
    }

    return ResultCode_Ok;
}

static ResultCode startIpc()
{
    // Set localId
    char procName[5] = "COREx";
    procName[4] = '0' + DNUM;
    MultiProc_setLocalId(MultiProc_getId(procName));

    // Start IPC
    int ret = Ipc_start();
    if (ret != Ipc_S_SUCCESS) {
        printf("Failed to initialize IPC (%d)\n", ret);
        return ResultCode_Error;
    }

    if (attachIpc() != ResultCode_Ok) {
        return ResultCode_Error;
    }

    return ResultCode_Ok;
}

static void initializeTask(UArg arg0, UArg arg1)
{
    puts("Configuring caching");

    initializeCaching();

    puts("Starting IPC");

    if (startIpc() != ResultCode_Ok) {
        goto INITIALIZE_END;
    }

    if (GlobalGates_init() != ResultCode_Ok) {
        puts("Failed to initialize global gates");
        goto INITIALIZE_END;
    }

    puts("Creating MSMC heap");

    if (createSharedHeap(&msmcHeap, msmcHeapRegionSize, "msmcHeap", 1, 1) != ResultCode_Ok) {
        goto INITIALIZE_END;
    }

    puts("Creating DDR3 heap");

    if (createSharedHeap(&ddr3Heap, ddr3HeapRegionSize, "ddr3Heap", 2, 2) != ResultCode_Ok) {
        goto INITIALIZE_END;
    }

    puts("Creating DDR3 uncached heap");

    if (createSharedHeap(&ddr3UncachedHeap, ddr3UncachedHeapRegionSize, "ddr3UncachedHeap", 3, 3) != ResultCode_Ok) {
        goto INITIALIZE_END;
    }

    puts("Initializing RMAN");

    if (initializeRman() != ResultCode_Ok) {
        goto INITIALIZE_END;
    }

    puts("Processing configuration");

    if (processConfiguration() != ResultCode_Ok) {
        goto INITIALIZE_END;
    }

    fflush(stdout);
    return;

INITIALIZE_END:

    fflush(stdout);
    BIOS_exit(0);
}

void main()
{
    const DspConfiguration* configuration = Configuration_get();

    if (configuration->marker != CONFIGURATION_MARKER) { // Wrong configuration
        goto END;
    }

    printf("main() on CORE=%d\n", DNUM);
    fflush(stdout);

    puts("Starting SYS/BIOS");
    fflush(stdout);

    Error_Block eb;
    Error_init(&eb);
    Task_Params taskParams;
    Task_Params_init(&taskParams);
    taskParams.stackSize = 0x1000;
    taskParams.arg0 = 0;
    taskParams.arg1 = 0;

    if (Task_create(initializeTask, &taskParams, &eb) == NULL) {
        puts("Failed to create initialization task");
        goto END;
    }

    BIOS_start();

END:

    fflush(stdout);
    BIOS_exit(0);
}
